import pymongo
import os, subprocess
import sys
import re


class MongoClient():
    '''
    Connects to mongo database and executes scripts
    '''
    def __init__(self, config_file, db_credentials, db_host, debug=False):

        self.config_file = config_file
        self.db_credentials = db_credentials
        self.db_host = db_host
        self.debug = debug

        return

    def get_db_conn(self, auth_db="admin"):

        mongo_credentials = self.db_credentials[self.db_host]
        mongo_host = mongo_credentials['host']
        mongo_port = mongo_credentials['port']
        mongo_user = mongo_credentials['user']
        mongo_password = mongo_credentials['password']

        try:
            mongo_client = pymongo.MongoClient(str(mongo_host), int(mongo_port))
        except pymongo.errors.ConnectionFailure as e:
            print ("Could not connect to server: {0}".format(e))
            raise Exception('Unable to connect to mongo host ' + mongo_host)

        if auth_db != "admin":
            try:
                mongo_auth_database = mongo_credentials['auth_database']
            except:
                mongo_auth_database = "admin"
        else:
            mongo_auth_database = auth_db

        try:
            conn = mongo_client[mongo_auth_database]
            conn.authenticate(mongo_user, mongo_password)
            self.mongo_client = mongo_client
        except pymongo.errors.ConnectionFailure as e:
            print ("authentication on auth_database {0} failed".format(mongo_auth_database))
            print ("Could not connect to server: {0}".format(e))
            mongo_client.close()
            raise Exception('Unable to authenticate using auth_database ' + mongo_auth_database)

        return self.mongo_client

    def close_connection(self):
        self.mongo_client.close()
        return

    def drop_collection(self, collection_name, database_name=""):

        mongo_credentials = self.db_credentials[self.db_host]
        if database_name == "":
            mongo_database = mongo_credentials['database']
        else:
            mongo_database = database_name
        result = self.mongo_client[mongo_database].drop_collection(collection_name)
        return result

    def set_mongo_collection(self, collection_name, database_name=""):

        mongo_credentials = self.db_credentials[self.db_host]
        if database_name == "":
            mongo_database = mongo_credentials['database']
        else:
            mongo_database = database_name
        self.collection = self.mongo_client[mongo_database][collection_name]
        return

    def create_collection_index(self, index_name, field_list=[], unique=False,
                                drop_index=True, default_language='english'):

        mongo_collection = self.collection
        index_field_list = []
        for field in field_list:
            for key, value in field.items():
                if value == "A":
                    field_tuple = (key, pymongo.ASCENDING)
                elif value == "D":
                    field_tuple = (key, pymongo.DESCENDING)
                index_field_list.append(field_tuple)

        if len(index_field_list) > 0:
            if drop_index:
                try:
                    mongo_collection.drop_index(index_name)
                except:
                    pass

            result = mongo_collection.create_index(index_field_list,
                                          name=index_name,
                                          unique=unique,
                                          default_language=default_language)
        return result

    def drop_collection_index(self, index_name):
        result = self.collection.drop_index(index_name)
        return result

    def reindex_collection(self):
        result = self.collection.reindex()
        return result

    def run_mongo_script_file(self, script_file, log_file, out_file='',
                              indicator_text="my results start from here"):

        mongo_credentials = self.db_credentials[self.db_host]
        mongo_user = mongo_credentials['user']
        mongo_password = mongo_credentials['password']
        mongo_host = mongo_credentials['host']
        mongo_port = mongo_credentials['port']
        mongo_database = mongo_credentials['database']
        mongo_auth_database = mongo_credentials['auth_database']

        ''' validate authentication '''
        mongo_client = self.get_db_conn(auth_db=mongo_auth_database)
        mongo_client.close()

        mongo_script_string = "mongo --host " + str(mongo_host) + \
                              " --port " + str(mongo_port) + \
                              " --authenticationDatabase " + str(mongo_auth_database) + \
                              " -u " + str(mongo_user) + \
                              " -p " + str(mongo_password) + \
                              " --quiet " + mongo_database + \
                              " < " + os.path.abspath(script_file)

        if out_file != '':
            mongo_script_string += " > " + os.path.abspath(out_file)

        if self.debug:
            print(mongo_script_string)

        output, error = subprocess.Popen(mongo_script_string, shell=True, stderr=subprocess.STDOUT,
                                         stdout=subprocess.PIPE).communicate()
        log_out = open(log_file, 'a')
        if sys.version_info.major == 3:
            output = output.splitlines()
            for line in output:
                log_out.write(str(line, 'utf-8'))
                print(str(line, 'utf-8'))
        else:
            log_out.write(output)
            print(output)

        if out_file != '':
            # Check the out file for any errors
            log_in = open(out_file, 'r')
            line_no = 0
            for log_line in log_in:
                line_no += 1
                error_position = log_line.find("errmsg")
                if error_position > 0:
                    print("Error while processing ")
                    print("Check line number " + str(line_no) + " in " + log_in.name)
                    log_in.close()
                    raise Exception('mongo script execution failed')
                    break
            log_in.close()

        # Check the log file for any errors
        log_in = open(log_file, 'r')
        line_no = 0
        for log_line in log_in:
            line_no += 1
            error_position = log_line.find("errmsg")
            if error_position > 0:
                print("Error while processing ")
                print("Check line number " + str(line_no) + " in " + log_in.name)
                log_in.close()
                raise Exception('mongo script execution failed')
                break
        log_in.close()

        # Generate final results file
        format_output = """ line_number=`cat {out_file} | grep -n "{text}" | cut -f1 -d':'` &&
        line_number=`expr $line_number + 1` && tail -n +$line_number {out_file} > $HOME/temp1.txt &&
        mv $HOME/temp1.txt {out_file}
        """.format(out_file=os.path.abspath(out_file), text=indicator_text)

        if self.debug:
            print(format_output)
        out, err = subprocess.Popen(format_output, shell=True, stderr=subprocess.STDOUT,
                                    stdout=subprocess.PIPE).communicate()
        if sys.version_info.major == 3:
            out = out.splitlines()
            for line in out:
                print(str(line, 'utf-8'))
        else:
            print(out)

        return

    def load_file_to_collection(self, collection_name, input_file, log_file,
                                database_name="", csv_file=False, drop_target=False):

        mongo_credentials = self.db_credentials[self.db_host]
        mongo_user = mongo_credentials['user']
        mongo_password = mongo_credentials['password']
        mongo_host = mongo_credentials['host']
        mongo_port = mongo_credentials['port']
        if database_name == "":
            mongo_database = mongo_credentials['database']
        else:
            mongo_database = database_name
        mongo_auth_database = mongo_credentials['auth_database']

        ''' validate authentication '''
        mongo_client = self.get_db_conn(auth_db=mongo_auth_database)
        mongo_client.close()

        mongo_import_string = "mongoimport --host " + str(mongo_host) + \
                              " --port " + str(mongo_port) + \
                              " --authenticationDatabase " + str(mongo_auth_database) + \
                              " -u " + str(mongo_user) + \
                              " -p " + str(mongo_password) + " -d " + str(mongo_database) + \
                              " -c " + collection_name

        if csv_file:
            mongo_import_string += " --type csv --headerline "

        mongo_import_string += " --file " + os.path.abspath(input_file)

        if drop_target:
            mongo_import_string += " --drop "

        if self.debug:
            print(mongo_import_string)

        output, error = subprocess.Popen(mongo_import_string, shell=True, stderr=subprocess.STDOUT,
                                         stdout=subprocess.PIPE).communicate()
        log_out = open(log_file, 'a')
        if sys.version_info.major == 3:
            output = output.splitlines()
            for line in output:
                log_out.write(str(line, 'utf-8'))
                print(str(line, 'utf-8'))
        else:
            log_out.write(output)
            print(output)

        # Check the log file for any errors
        log_in = open(log_file, 'r')
        line_no = 0
        for log_line in log_in:
            line_no += 1
            error_position = log_line.find("Failed")
            if error_position > 0:
                print("Error while processing ")
                print("Check line number " + str(line_no) + " in " + log_in.name)
                log_in.close()
                raise Exception('import to mongodb failed - Error while running mongoimport')
                break
        log_in.close()

        return

    def export_collection_to_file(self, collection_name, output_file, log_file,
                                  database_name="", csv_file=False, csv_fields=""):

        mongo_credentials = self.db_credentials[self.db_host]
        mongo_user = mongo_credentials['user']
        mongo_password = mongo_credentials['password']
        mongo_host = mongo_credentials['host']
        mongo_port = mongo_credentials['port']
        if database_name == "":
            mongo_database = mongo_credentials['database']
        else:
            mongo_database = database_name
        mongo_auth_database = mongo_credentials['auth_database']

        ''' validate authentication '''
        mongo_client = self.get_db_conn(auth_db=mongo_auth_database)
        mongo_client.close()

        mongo_export_string = "mongoexport --host " + str(mongo_host) + \
                              " --port " + str(mongo_port) + \
                              " --authenticationDatabase " + str(mongo_auth_database) + \
                              " -u " + str(mongo_user) + \
                              " -p " + str(mongo_password) + " -d " + str(mongo_database) + \
                              " -c " + collection_name

        if csv_file:
            mongo_export_string += " --type=csv --fields " + csv_fields

        mongo_export_string += " --out " + os.path.abspath(output_file)

        if self.debug:
            print(mongo_export_string)

        output, error = subprocess.Popen(mongo_export_string, shell=True, stderr=subprocess.STDOUT,
                                         stdout=subprocess.PIPE).communicate()
        log_out = open(log_file, 'a')
        if sys.version_info.major == 3:
            output = output.splitlines()
            for line in output:
                log_out.write(str(line, 'utf-8'))
                print(str(line, 'utf-8'))
        else:
            log_out.write(output)
            print(output)

        # --------------------------------------------------------------------------------
        # Check the log file for any errors
        log_in = open(log_file, 'r')
        line_no = 0
        for log_line in log_in:
            line_no += 1
            error_position = log_line.find("Failed")
            if error_position > 0:
                print("Error while processing ")
                print("Check line number " + str(line_no) + " in " + log_in.name)
                log_in.close()
                raise Exception('export from mongodb failed - Error while running mongoexport')
                break
        log_in.close()

        return

    def export_mixpanel_query_results(self, collection_name, mixpanel_query, mixpanel_fields,
                                      results_csv_file, constant_fields={}, field_mapping={},
                                      file_delimiter=",", header_row=False):

        mongo_credentials = self.db_credentials[self.db_host]
        mongo_database = mongo_credentials['database']
        mongo_auth_database = mongo_credentials['auth_database']

        ''' validate authentication '''
        mongo_client = self.get_db_conn(auth_db=mongo_auth_database)

        ''' initialize collection '''
        mongo_collection = mongo_client[mongo_database][collection_name]

        with open(results_csv_file, 'w') as f:

            ''' header row '''
            if header_row:
                out_header_row = ""
                for config_data in field_mapping:
                    gp_column_name = config_data['gp_column_name']
                    out_header_row += str(gp_column_name) + str(file_delimiter)

                f.write(out_header_row.rstrip(str(file_delimiter)) + '/n')

            ''' detail row - process each row in the order in which the fields appear '''
            cursor = mongo_collection.find(mixpanel_query, mixpanel_fields)
            key_list = mixpanel_fields.keys()

            for row in cursor:

                out_dict = dict()
                out_dict["id"] = row["_id"]
                for key in key_list:
                    if key.startswith("properties"):
                        property_name = key.split('.')[1]
                        try:
                            field_value = row["properties"][property_name]
                        except KeyError:
                            field_value = ""
                    else:
                        field_value = row[key]

                    try:
                        out_dict[str(key)] = field_value.replace(file_delimiter, "").replace('\r\n', '').replace('\r', '').replace('\n', '')
                    except AttributeError:
                        out_dict[str(key)] = field_value

                out_row = ""
                for config_data in field_mapping:

                    gp_column_name = config_data['gp_column_name']
                    gp_column_position = config_data['gp_column_position']
                    gp_column_type = config_data['gp_column_type']
                    json_field_name = config_data['json_field']

                    if config_data['gp_value_is_constant']:
                        try:
                            column_value = constant_fields[gp_column_name]
                        except KeyError:
                            column_value = ""
                    else:
                        try:
                            column_value = out_dict[json_field_name]
                        except KeyError:
                            column_value = ""

                    if gp_column_type == "varchar":
                        try:
                            column_value = str(column_value)
                        except UnicodeError:
                            column_value = re.sub(r'[^\x00-\x7f]', r'', column_value)
                    else:
                        column_value = str(column_value)

                    if column_value == "None":
                        column_value = ""

                    if gp_column_position == 1:
                        out_row += column_value
                    else:
                        out_row += file_delimiter + column_value

                f.write(str(out_row) + "\n")

            f.close
            mongo_client.close()

        return




'''
This function can be used by any DAG that needs a python operator to run a sql file
'''
import datetime as dt
import json
import os
from random import randint
import pandas as pd
from dattasa import data_pipeline as dp


def run_script(**kwargs):
    '''
    The function takes in
        file name, db_host, db_config_file, audit_table_name
        sql path, output path, log path
        delimited out indicator and db host as input
    Default db_config_file = $HOME/database.yaml
    Default folders if not specified are $SQL_DIR, $TEMP_DIR, $LOG_DIR
    The function then searches for the db_host name in database.yaml file
    and gets the login credentials. The sql is run and the output and log files are redirected to the
    path specified. In case of any sql failures, the dattasa package will throw and exception and
    this function will not complete successfully

    Naming convention for the files
        sql_file name should be under <sql_path>/<file_name>.sql
        out_file name will be under <out_path>/<file_name>.out
        log_file name will be under <log_path>/<file_name>_'YYYY-MM-DD_HH:MM:SS'.log

    ADDTIONAL FUNCTIONALITY -
    pre_check_tables : Provide list of tables that have to be checked before running sql
    pre_check_scripts : Provide list of sql scripts that have to be checked before running sql
    load_tables : Provide list of tables that have to be audited as part of this run

    run_sql_file - By default value is True. If this is set to False the sql file will not be run. However you
    can still use this python function for testing pre-checks and/or auditing load tables

    SQL
    :param kwargs:
    :return:
    '''

    # Passing arguments to a function
    file_name = kwargs['file_name']
    audit_table = kwargs['audit_table']
    db_host = kwargs['db_host']

    db_config_file = kwargs.get('db_config_file', os.environ['HOME'] + '/database.yaml')
    sql_path = kwargs.get('sql_path', os.environ['SQL_DIR'])
    out_path = kwargs.get('out_path', os.environ['TEMP_DIR'])
    log_path = kwargs.get('log_path', os.environ['LOG_DIR'])

    pre_check_tables = kwargs.get('pre_check_tables', [])
    pre_check_scripts = kwargs.get('pre_check_scripts', [])
    load_tables = kwargs.get('load_tables', [])
    run_sql_file = kwargs.get('run_sql_file', True)

    delimited_file = kwargs.get('delimited_file', 'F')

    # Converting string to boolean value
    if delimited_file == 'T':
        delimited=True
    elif delimited_file == 'F':
        delimited=False

    '''
    Set all the files needed before calling run_psql_file function from greenplum_client
    You can either use os environment variables eg:- os.environ['HOME'] or airflow variables 
    '''
    sql_file = sql_path + '/' + file_name + '.sql'
    out_file = out_path + '/' + file_name + ".out"
    log_file = log_path + '/' + file_name + "_" + \
               dt.datetime.now().strftime("%Y-%m-%d_%H:%M:%S") + '.log'

    '''
    Perform pre-checks before running sql
    '''
    print ("Running pre-checks before running sql file " + file_name + "\n")
    run_pre_check_scripts(log_file, db_config_file, db_host, pre_check_tables, pre_check_scripts)

    audit_id = str(long(dt.datetime.now().strftime("%Y%m%d%H%M%S%f")) * 100 + randint(0, 9) * 10 + randint(0, 9))

    '''
    Create an entry in the audit table before running the sql for each load_name
    '''
    kwargs = {"load_type": "pre", "audit_id": audit_id, "audit_table": audit_table,
              "load_tables": load_tables, "load_description": file_name}
    load_audit_table(**kwargs)

    '''
    If sql file has to be run (Default=True) then run the sql file here.
    Note the sql will run only after all pre-checks succeed and entry in audit table is made succesfully
    '''
    if run_sql_file:
            db = dp.DataComponent().set_credentials(db_host, db_config_file)
            db.run_psql_file(sql_file, log_file, out_file, delimited)

    '''
    Update the audit table after sql run has completed and update final row counts for each load name
    '''
    kwargs = {"load_type": "post", "audit_id": audit_id, "audit_table": audit_table,
              "load_tables": load_tables, "load_description": file_name}
    load_audit_table(**kwargs)

    return


def run_pre_check_scripts(log_file, yaml_file, db_host, pre_check_tables, pre_check_scripts):
    '''

    :param log_file:
    :param yaml_file:
    :param db_host:
    :param pre_check_tables:
    :param pre_check_scripts:
    :return:
    '''

    db = dp.DataComponent().set_credentials(db_host, yaml_file)
    conn = db.get_db_conn(True, True, 20)

    out_log = open(log_file, "w")

    '''
    Ensure none of the pre check tables are empty.
    We will ask the DB for five rows and log whether we got them.
    If there were zero rows, fail the job.
    '''
    for table in pre_check_tables:
        pre_check_sql = "SELECT * FROM " + table + " LIMIT 5;"
        pre_check_df = pd.read_sql(pre_check_sql, conn)
        pre_check_row_count = pre_check_df.shape[0]
        out_log.write(str(pre_check_row_count) + " rows returned from " + table + "\n")
        if pre_check_row_count == 0:
            print ("No rows found in pre check table : " + table + "\n")
            out_log.close()
            db.close_connection()
            raise Exception('pre-check found an empty table ' + table)
            break

    '''
    Ensure none of the pre check sql scripts return empty result
    We will ask the DB for five rows and log whether we got them.
    If there were zero rows, fail the job.
    '''
    for sql_script in pre_check_scripts:
        pre_check_df = pd.read_sql(sql_script, conn)
        pre_check_row_count = pre_check_df.shape[0]
        out_log.write(str(pre_check_row_count) + " rows returned from " + sql_script + "\n")
        if pre_check_row_count == 0:
            print ("No rows found for pre check sql : " + sql_script + "\n")
            print ("SQL File " + file_name + " will not be  run. Aborting ..." + "\n")
            out_log.close()
            db.close_connection()
            raise Exception('pre-check did not return any rows for sql ' + sql_script)
            break

    out_log.close()
    db.close_connection()

    return


'''
        Example on how to call load_audit_table independently in another python program

import sys, os
import datetime as dt
from random import randint
from data_jedi import dag_controller

python_file=os.environ['AIRFLOW_HOME'] + '/dags'
sys.path.append(python_file)
import dag_controller
load_tables=['adm.adm.fact_employer_feature', 'adm.fact_account_feature']
load_desc="reg_rate_tracking"
audit_id = str(long(dt.datetime.now().strftime("%Y%m%d%H%M%S%f")) * 100 + randint(0, 9) * 10 + randint(0, 9))

kwargs={"load_type": "pre", "audit_id": audit_id, "audit_table": "adm.airflow_load_audit",
"load_tables": load_tables, "load_description": load_desc}
dag_controller.load_audit_table(**kwargs)
kwargs={"load_type": "post", "audit_id": audit_id, "audit_table": "adm.airflow_load_audit",
"load_tables": load_tables, "load_description": load_desc}
dag_controller.load_audit_table(**kwargs)

'''


def load_audit_table(**kwargs):
    '''
    :param load_type:
    :param load_tables:
    :param load_description:
    :return:
    '''

    audit_id = kwargs['audit_id']     # Required Parameter
    load_type = kwargs['load_type']   # Required Parameter
    audit_table = kwargs['audit_table'] # Required Parameter
    queue_name = kwargs['queue_name'] # Required Parameter
    rabbitmq_host = kwargs['rabbitmq_host']  # Required Parameter

    load_tables = kwargs.get('load_tables', [])
    load_description = kwargs.get('load_description', '')
    rabbitmq_config_yaml = kwargs.get('config_yaml', os.environ['HOME'] + '/database.yaml')

    producer = dp.DataComponent().set_credentials(rabbitmq_host, rabbitmq_config_yaml)

    ''' Publish a message for each table '''
    for table_name in load_tables:
        data = {
            "audit_id": audit_id,
            "load_type": load_type,
            "load_tables": table_name,
            "load_description": load_description,
            "audit_table": audit_table,
            "load_time": dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        }
        message = json.dumps(data)
        status = producer.send(queue_name, message, auto_delete=False)
        print("published message for creating " + load_type + "-load audit record for " + table_name)

    producer.stop()

    return


def get_copy_mysql_table_to_gp_command(source_table, target_table, source_conn='',
                                       target_conn='', error_limit='2', null_string='',
                                       description=''):
    if null_string != '':
        load_command = 'python ' + os.environ['PROJECT_HOME'] + '/python_bash_scripts copy_mysql_table_to_gp.py ' + \
                       ' -s ' + source_conn + ' -t ' + target_conn + ' -x ' + source_table + \
                       ' -y ' + target_table + ' -d ' + description + ' -e ' + error_limit + \
                       ' -n ' + null_string
    else:
        load_command = 'python ' + os.environ['PROJECT_HOME'] + '/python_bash_scripts copy_mysql_table_to_gp.py ' + \
                       ' -s ' + source_conn + ' -t ' + target_conn + ' -x ' + source_table + \
                       ' -y ' + target_table + ' -d ' + description + ' -e ' + error_limit

    return load_command


def get_copy_postgres_table_to_gp_command(source_table, target_table, source_conn='',
                                          target_conn='', error_limit='2', null_string='',
                                          description=''):

    if null_string != '':
        load_command = 'python ' + os.environ['PROJECT_HOME'] + '/python_bash_scripts copy_postgres_table_to_gp.py ' + \
                       ' -s ' + source_conn + ' -t ' + target_conn + ' -x ' + source_table + \
                       ' -y ' + target_table + ' -d ' + description + ' -e ' + error_limit + \
                       ' -n ' + null_string
    else:
        load_command = 'python ' + os.environ['PROJECT_HOME'] + '/python_bash_scripts copy_postgres_table_to_gp.py ' + \
                       ' -s ' + source_conn + ' -t ' + target_conn + ' -x ' + source_table + \
                       ' -y ' + target_table + ' -d ' + description + ' -e ' + error_limit

    return load_command



from dattasa.data_pipeline import DataComponent
from dattasa.data_pipeline import APICall
from dattasa.data_pipeline import DataProcessor
from dattasa.file_processor import File
from dattasa.file_processor import FileTransfer




import paramiko
import sys
import yaml
import os
import gnupg
import pprint
import json
import csv


class File():

    def __init__(self, file_name):
        self.file = open(file_name, 'r')
        return

    def encrypt(self, gpg_key_user, gpg_output_file, log_file, gnupg_home="", verbose=False):

        log_out = open(log_file, 'a')
        gpg_fingerprint = ""
        if gnupg_home == '':
            gnupg_home = os.environ['HOME'] + '/.gnupg'
        gpg = gnupg.GPG(gnupghome=gnupg_home, verbose=verbose)
        gpg_list = gpg.list_keys()
        for gpg_key in gpg_list:
            if verbose:
                pprint.pprint(gpg_key)
            if gpg_key_user in gpg_key['uids']:
                gpg_fingerprint = gpg_key['fingerprint']
                break

        if gpg_fingerprint != "":
            status = gpg.encrypt_file(self.file, gpg_fingerprint, output=gpg_output_file)

            if verbose:
                print ('ok: ' + str(status.ok))
                print ('status: ' + str(status.status))

            if not status.ok:
                '''  run this command manually if needed to troubleshoot '''
                gpg_command = "gpg --encrypt --recipient " + gpg_key_user + " " + self.file.name + \
                              " && mv " + self.file.name + ".gpg " + gpg_output_file
                log_out.write(gpg_command)
                log_out.write("run this command manually if needed to troubleshoot" + "\n")
                log_out.write(gpg_command + "\n")
                log_out.write("Encyrption failed. Status : " + str(status.status) + "\n")
                self.file.close()
                log_out.close()
                raise Exception('Encyrption failed. Status : ' + str(status.status))
        else:
            log_out.write("No valid gpg key found for " + gpg_key_user + " \n")
            self.file.close()
            log_out.close()
            raise Exception('No valid gpg key found for ' + gpg_key_user)

        return open(gpg_output_file, 'r')

    def row_count(self):

        try:
            with self.file as in_file:
                count = sum(1 for _ in in_file)
            self.file.close()
        except:
            print (sys.exc_info())
            self.file.close()
            raise Exception('Unable to obatin row count')

        return count

    def csv_to_fixed_width(self, out_file_name, input_header_row=False, input_delimiter='|',
                           output_field_length=[], output_field_alignment=[]):

        try:
            with self.file as in_file, open(out_file_name, "w") as output_file:
                line_no = 1
                for line in in_file:
                    ''' convert each csv line to fixed width line '''
                    out_line = ''
                    word_list = line.split(input_delimiter)
                    position = 0
                    for word in word_list:
                        ''' ljust or rjust depends on output_field_alignment value of 1 or 0
                        header row is always left justified '''
                        if output_field_alignment[position] == 1 or (line_no == 1 and input_header_row):
                            ''' left justify '''
                            out_line += word.strip().ljust(output_field_length[position])
                        else:
                            ''' right justify '''
                            out_line += word.strip().rjust(output_field_length[position])
                        position += 1
                    output_file.write(out_line + '\n')
                    line_no += 1

                ''' Finally close the file '''
                output_file.close()

        except:
            print (sys.exc_info())
            self.file.close()
            output_file.close()
            raise Exception('Unable to convert to fixed width')

        return

    def replace_words(self, word_replacements={}, inplace=True, output_file_name=''):

        backup_file_name = self.file.name + '.bkup'
        with open(backup_file_name, "w") as output_file, self.file as input_file:
            for line in input_file:
                if sys.version_info.major == 2:
                    for key, value in word_replacements.iteritems():
                        line = line.replace(key, value)
                else:
                    for key, value in word_replacements.items():
                        line = line.replace(key, value)
                output_file.write(line)
            input_file.close()
            output_file.close()

        if inplace:
            os.rename(backup_file_name, self.file.name)
        else:
            os.rename(backup_file_name, output_file_name)

        ''' Open the file again '''
        self.file = open(self.file.name, 'r')

        return


class FileTransfer():

    def __init__(self, sftp_site, config_file=''):
        if config_file == '':
            self.config_file = os.environ['HOME'] + "/ftpsites.yaml"
        else:
            self.config_file = config_file

        try:
            from yaml import CLoader as Loader , CDumper as Dumper
        except ImportError:
            from yaml import Loader, Dumper

        with open(config_file , 'r') as f:
            file_credentials = yaml.load(f, Loader=Loader)
            self.sftp_site = sftp_site
            self.sftp_credentials = file_credentials[sftp_site]

        return

    def sftp_send_file(self, local_file_path, local_file_list, log_file,
                       remote_file_list='', encrypted_password=False):

        sftp_credentials = self.sftp_credentials
        sftp_site = self.sftp_site
        host = sftp_credentials['host']
        port = sftp_credentials['port']
        username = sftp_credentials['user']
        remote_file_path = sftp_credentials['folder']
        if not encrypted_password:
            password = sftp_credentials['password']

        ''' Create list of files to be sent '''
        local_file_list = local_file_list.split(',')
        remote_file_list = remote_file_list.split(',')

        log_out = open(log_file, 'a')
        log_out.write("\n\nstarting sftp to site : " + sftp_site + "\n")
        if remote_file_path is not None:
            log_out.write("sftp to remote_file_path : " + remote_file_path + "\n")
            remote_file_path += "/"
        else:
            log_out.write("sftp to default home directory \n")
            remote_file_path = "/"
        transport = paramiko.Transport((host, port))
        transport.connect(username=username, password=password)

        file_index = 0
        for local_file_name in local_file_list:

            try:
                remote_file_name = remote_file_list[file_index]
            except:
                remote_file_name = local_file_name

            try:
                log_out.write("sending file: " + local_file_name + "\n")

                sftp = paramiko.SFTPClient.from_transport(transport)
                remote_file = remote_file_path + remote_file_name
                local_file = local_file_path + local_file_name
                sftp.put(local_file, remote_file)
                sftp.close()
                transport.close()
                log_out.write("sftp put complete!! \n")
                log_out.close()

            except OSError as err:
                log_out.write("ERROR: OS error - " + format(err) + "\n")
                log_out.close()
                print ("ERROR: OS error - " + format(err) + "\n")
            except:
                log_out.write("ERROR: Unexpected error while sending " + str(sys.exc_info()[0]) + "\n")
                log_out.close()
                print ("ERROR: Unexpected error - ")
                print (sys.exc_info())

            file_index += 1

        return


def _byteify(data, ignore_dicts=False):
    # if this is a unicode string, return its string representation
    if sys.version_info.major == 2:
        if isinstance(data, unicode):
            return data.encode('utf-8')
    # if this is a list of values, return list of byteified values
    if isinstance(data, list):
        return [_byteify(item, ignore_dicts=True) for item in data]
    # if this is a dictionary, return dictionary of byteified keys and values
    # but only if we haven't already byteified it
    if isinstance(data, dict) and not ignore_dicts:
        return {
            _byteify(key, ignore_dicts=True): _byteify(value, ignore_dicts=True)
            for key, value in data.items()
        }
    # if it's anything else, return it in its original form
    return data


def json_load_byteified(file_handle):
    return _byteify(
        json.load(file_handle, object_hook=_byteify) ,
        ignore_dicts=True
    )


def json_loads_byteified(json_text):
    return _byteify(
        json.loads(json_text, object_hook=_byteify),
        ignore_dicts=True
    )


def write_dict_to_csv(csv_file, csv_columns, dict_data, delimiter):
    try:
        with open(csv_file , 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=csv_columns, delimiter=delimiter)
            writer.writeheader()
            for data in dict_data:
                writer.writerow(data)
    except IOError:
        print("I/O error({0}): {1}".format(IOError.errno, IOError.strerror))
    return



import pymysql
import sqlalchemy
from sqlalchemy import exc
import time
import datetime as dt
import subprocess
import sys, os


class MySQLClient():
    '''
    Connects to mysql database and executes sql queries against db
    '''
    def __init__(self, config_file, db_credentials, db_host):
        self.config_file = config_file
        self.db_credentials = db_credentials
        self.db_host = db_host
        return

    def get_db_conn(self, use_pymysql=True, verbose=False, connect_timeout=10):
        '''
        uses the credentials in config_yaml_file for db_host while initiating database connection
        connection can be established by either using pyscopg2 or sqlachemy modules
        :param config_yaml_file:
        :param db_host:
        :param use_pyscopg2:
        :param verbose:
        :param connect_timeout:
        :param use_ssl:
        :param ssl_params:
        :return: database connection session that has been made using thecredentials
        '''

        db_credentials = self.db_credentials
        db_host = self.db_host

        try:
            ssl_params = db_credentials[db_host]['ssl_params']
            use_ssl = True
        except KeyError:
            use_ssl = False

        if use_pymysql:
            try:
                if use_ssl:
                    self.conn = pymysql.connect(host=db_credentials[db_host]['host'],
                                                port=db_credentials[db_host]['port'],
                                                user=db_credentials[db_host]['user'],
                                                ssl=ssl_params,
                                                password=db_credentials[db_host]['password'],
                                                connect_timeout = connect_timeout)
                else:
                    self.conn = pymysql.connect(host=db_credentials[db_host]['host'] ,
                                                port=db_credentials[db_host]['port'] ,
                                                user=db_credentials[db_host]['user'] ,
                                                password=db_credentials[db_host]['password'] ,
                                                connect_timeout=connect_timeout)
                if verbose:
                    print("db credentials import succeeded. Connected using pymysql ..")

            except pymysql.MySQLError as e:
                print("Unable to connect using pymysql")
                print(e, e.args)
            except Exception as e:
                print(e, e.args)

        else:
            # Using sqlalchemy to make the connection
            try:
                connect_string = "mysql://" + db_credentials[db_host]['user'] + ":" + \
                                 db_credentials[db_host]['password'] + \
                                 "@" + db_credentials[db_host]['host']
                engine = sqlalchemy.create_engine(connect_string)
                c = engine.connect()
                self.conn = c.connection
                if verbose:
                    print("db credentials import succeeded. connected successfully using sqlalchemy ..")

            except exc.SQLAlchemyError as e:
                print("Unable to connect using sqlalchemy!")
            except Exception as e:
                print(e, e.args)

        return self.conn

    def get_db_cursor(self, verbose=False, connect_timeout=10):
        '''
        uses the credentials in config_yaml_file for db_host while initiating database connection.
        Return the database cursor after making the database connection
        :param config_yaml_file:
        :param db_host:
        :param verbose:
        :param connect_timeout:
        :return:
        '''

        db_credentials = self.db_credentials
        db_host = self.db_host

        try:
            self.conn = pymysql.connect(host=db_credentials[db_host]['host'],
                                        port=db_credentials[db_host]['port'],
                                        user=db_credentials[db_host]['user'],
                                        password=db_credentials[db_host]['password'],
                                        connect_timeout=connect_timeout)
            if verbose:
                print("db credentials import succeeded. Connected using pymysql ..")

        except KeyError:
            print("No password found in yaml file. Using password from pgpass")
            self.conn = pymysql.connect(user=db_credentials[db_host]['user'],
                                        host=db_credentials[db_host]['host'],
                                        port=db_credentials[db_host]['port'],
                                        connect_timeout=connect_timeout)
            if verbose:
                print("db credentials import succeeded. Connected using pymysql ..")

        except pymysql.Error as e:
            print("Unable to connect using pymysql!")
            print(e, e.args)
        except Exception as e:
            print(e, e.args)

        try:
            self.cur = self.conn.cursor()
            if verbose:
                print("opened database successfully")
        except Exception as e:
            print("unable to connect to database", "error message: {0}".format(e))

        return self.cur

    def execute_query(self, query):
        try:
            self.cur.execute(query)
        except:
            self.conn.rollback()
            raise
        return None

    def commit_query(self):
        try:
            self.conn.commit()
            print("successful commit")
        except Exception as e:
            print("error message: {0}".format(e))

    def rollback(self):
        try:
            self.conn.rollback()
            print("successful rollback")
        except Exception as e:
            print("error message: {0}".format(e))

    def get_results(self):
        return self.cur.fetchall()

    def close_connection(self):
        self.conn.close()
        return None

    def run_sql_command(self, sql_query, fetch=False):
        '''
        The function takes a filename as input
        and will run the SQL query using previously established connection
        pyscopg2 connection must be established prior to calling this function
        :param sql_query_file:
        :return:
        '''

        try:
            start = time.time()
            print("Started sql: " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            cursor = self.conn.cursor()
            cursor.execute(sql_query)
            self.conn.commit()

            if fetch:
                results = cursor.fetchall()
            else:
                results = None

            cursor.close()

            end = time.time()
            print("Ended sql: " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            print("Overall run time:")
            print(str((end - start) * 1000) + ' ms')

        except pymysql.MySQLError as e:
            print("Unable to connect using pymysql!")
            print(e, e.args)
        except OSError as err:
            print("OS error: " + format(err))
            cursor.close()
        except Exception as e:
            print("error message: {0}".format(e))
            cursor.close()

        return results

    def run_sql_file(self, sql_query_file, fetch=False):
        '''
        The function takes a filename as input
        and will run the SQL query using previously established connection
        pyscopg2 connection must be established prior to calling this function
        :param sql_query_file:
        :return:
        '''

        try:
            f = open(sql_query_file, 'r')
            start = time.time()
            sql = " ".join(f.readlines())

            print(
                "Started executing: " + f.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            cursor = self.conn.cursor()
            cursor.execute(sql)
            self.conn.commit()

            if fetch:
                results = cursor.fetchall()
            else:
                results = None

            cursor.close()

            end = time.time()
            print("Finished executing: " + f.name + " at " + str(
                dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            print("Overall run time:")
            print(str((end - start) * 1000) + ' ms')

        except pymysql.MySQLError as e:
            print("Unable to connect using pymysql!")
            print(e, e.args)
        except OSError as err:
            print("OS error: " + format(err))
            cursor.close()
        except Exception as e:
            print("error message: {0}".format(e))
            cursor.close()

        return results

    def export_sql_results(self, sql_file, log_file, output_file,
                           output_file_delimited=False, delimiter=','):

        sql_in = open(sql_file, 'r')
        results_out = open(output_file, 'a')
        log_out = open(log_file, 'a')
        db_credentials = self.db_credentials
        db_group_suffix = self.db_host

        ''' Create ~/.my.cnf file '''
        config_out = open(os.environ['HOME'] + '/.my.cnf', 'w')
        config_out.write("[client" + db_group_suffix + "]" + "\n")
        config_out.write("user = " + db_credentials[db_group_suffix]['user'] + "\n")
        config_out.write("password = " + db_credentials[db_group_suffix]['password'] + "\n")
        config_out.write("host = " + db_credentials[db_group_suffix]['host'] + "\n")
        config_out.write("port = " + str(db_credentials[db_group_suffix]['port']) + "\n")
        config_out.close()

        ''' Set Permissions '''
        bash_command = "chmod 600 $HOME/.my.cnf"
        output, error = subprocess.Popen(bash_command, shell=True, stderr=subprocess.STDOUT,
                                         stdout=subprocess.PIPE).communicate()
        if sys.version_info.major == 3:
            output = output.splitlines()
            for line in output:
                log_out.write(str(line, 'utf-8'))
        else:
            log_out.write(output)

        try:
            start = time.time()
            print ("Started: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            log_out.write(
                "Started: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            mysql_string = """ mysql --defaults-group-suffix={db_group_suffix} --connect_timeout=7200 -A < {sql_file} > {output_file}
                            """.format(db_group_suffix=db_group_suffix, sql_file=os.path.abspath(sql_in.name),
                                       output_file=os.path.abspath(results_out.name))

            output, error = subprocess.Popen(mysql_string, shell=True, stderr=subprocess.STDOUT,
                                             stdout=subprocess.PIPE).communicate()

            if sys.version_info.major == 3:
                output = output.splitlines()
                for line in output:
                    log_out.write(str(line, 'utf-8'))
            else:
                log_out.write(output)

            end = time.time()
            print ("Ended: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            log_out.write("Ended: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            total_time = (end - start) * 1000
            print ("Total time taken to run: " + str(total_time) + " ms \n")
            log_out.write("Total time taken to run: " + str(total_time) + " ms \n")

            results_out.close()
            log_out.close()

            if output_file_delimited:
                subs_command = """sed -i -e 's/\t/""" + delimiter + """/g'  """ + os.path.abspath(results_out.name)
                subprocess.Popen(subs_command, shell=True, stderr=subprocess.STDOUT)

        except OSError as err:
            print("OS error: {0)".format(err))
            log_out.write("OS error: {0)".format(err))
        except Exception as e:
            print("error message: {0}".format(e))
            log_out.write("error message: {0}".format(e))

        # --------------------------------------------------------------------------------
        # Check the sql output for any errors
        error_position = 0
        if sys.version_info.major == 3:
            for line in output:
                out_line = str(line, 'utf-8')
                error_position = out_line.find('error')
                if error_position > 0:
                    break
        else:
            error_position = output.find('error', 0)

        if error_position > 0:
            print ("Error while processing " + sql_in.name)
            print ("**************************** START OF LOG *******************************")
            print(output)
            print ("***************************** END OF LOG ********************************")
            raise Exception('mysql execution failed - Error while running sql')

        return

    def load_csv_to_table(self, data_file, delimiter, schema_name, table_name):
        '''
        The function takes input file, file delimiter and a connection as input
        and loads the data into the table. File has to be in csv format and delimited.
        target table must be empty or will be truncated prior to load
        pymysql connection must be established prior to calling this function
        :param data_file:
        :param delimiter:
        :param conn:
        :param schema_name:
        :param table_name:
        :return:
        '''

        try:
            start = time.time()
            table = schema_name + '.' + table_name
            print("Started load: " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            cursor = self.conn.cursor()
            load_command = """
            LOAD DATA INFILE '{data_file}'
            INTO TABLE {table}
            FIELDS TERMINATED BY '{delimiter}'
            ENCLOSED BY '"'
            LINES TERMINATED BY '\r\n'
            IGNORE 1 LINES;
            """.format(data_file=data_file, table=table, delimiter=delimiter)
            cursor.execute(load_command)
            self.conn.commit()
            cursor.close()
            end = time.time()
            print("Ended load: " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            print("Overall run time:")
            print(str((end - start) * 1000) + ' ms')

        except pymysql.MySQLError as e:
            print("Unable to connect using pymysql!")
            print(e, e.args)
        except OSError as err:
            print("OS error: " + format(err))
            cursor.close()
        except Exception as e:
            print("error message: {0}".format(e))
            cursor.close()

        return



from simple_salesforce import Salesforce
import sys


class SalesforceClient():
    def __init__(self, api_credentials):
        self.api_credentials = api_credentials

        try:
            sfdc_client = Salesforce(username=api_credentials['salesforce_user'],
                                     password=api_credentials['salesforce_password'],
                                     security_token=api_credentials['apisecret'])
            self.sfdc_client = sfdc_client

        except Exception as e:
            print(e, e.args)
            raise Exception("Unable to connect to salesforce")

        return

    def request_data(self, object_name, field_list=[], query_condition='', query_all=False, debug=False):

        if len(field_list) == 0:
            raise Exception("field_list is empty. cannot export results to csv")
        else:
            output_fields = ','.join(field_list)

        sfdc_query = "SELECT " + output_fields + " FROM " + object_name
        if query_condition != '':
            sfdc_query += " WHERE " + query_condition

        if debug:
            print("Fetching results for query - \n")
            print(sfdc_query)

        query_result = self.sfdc_client.query(sfdc_query)
        record_list = query_result['records']  # dictionary of results!
        json_response = []
        while len(record_list) > 0:
            for record in record_list:
                message = {}
                for field_lkp in field_list:
                    field_name = field_lkp.split('.')
                    first_name = field_name[0]
                    try:
                        second_name = field_name[1]
                    except IndexError:
                        second_name = ''

                    if record[first_name] is None:
                        attr_value = ""
                    else:
                        if second_name != '':
                            attr_value = record[first_name][second_name]
                        else:
                            attr_value = record[first_name]

                    if sys.version_info >= (3, 0, 0):
                        # for Python 3
                        if isinstance(attr_value, bytes):
                            attr_value = attr_value.decode('ascii')  # or  s = str(s)[2:-1]
                    else:
                        # for Python 2
                        if type(attr_value) == unicode:
                            attr_value = attr_value.encode('utf-8')

                    message[field_lkp] = attr_value
                json_response.append(message)

            try:
                url = query_result['nextRecordsUrl']
                if debug:
                    print(url)

                if query_all:
                    query_result = self.sfdc_client.query_more(url, True)
                    record_list = query_result['records']  # dictionary of results!
                else:
                    print (
                        "WARNING - Resultset exceeds 2000 rows. Stopping at 2000 rows. Use query_all=True to get all rows")
                    record_list = []
            except KeyError:
                record_list = []

        return json_response

    def export_query_results_to_csv(self, object_name, output_file_name, field_list=[],
                                    query_condition='', file_delimiter=',', debug=False):

        if len(field_list) == 0:
            raise Exception("field_list is empty. cannot export results to csv")
        else:
            out_field_list = []
            for attr_value in field_list:
                out_field_list.append(attr_value['field'])
            output_fields = ','.join(out_field_list)

        sfdc_query = "SELECT " + output_fields + " FROM " + object_name
        if query_condition != '':
            sfdc_query += " WHERE " + query_condition

        if debug:
            print("Fetching results for query - \n")
            print(sfdc_query)

        try:
            query_result = self.sfdc_client.query(sfdc_query)
            record_list = query_result['records']  # dictionary of results!
        except Exception as e:
            print(e, e.args)
            raise Exception("Unable to fetch results from salesforce")

        sfdc_file = open(output_file_name , 'w')
        try:
            for record in record_list:

                row = ''
                for field_lkp in field_list:
                    field_name = field_lkp['field']
                    field_name = field_name.split('.')
                    first_name = field_name[0]
                    try:
                        second_name = field_name[1]
                    except IndexError:
                        second_name = ''

                    if record[first_name] is None:
                        row += ''
                    else:
                        if second_name != '':
                            attr_value = record[first_name][second_name]
                        else:
                            attr_value = record[first_name]

                        ''' remove non-ascii characters by ignoring errors '''
                        if sys.version_info.major == 2:
                            if type(record[first_name]) == unicode:
                                attr_value = str(attr_value.encode("utf-8", errors="ignore"))
                            else:
                                attr_value = str(attr_value)
                        else:
                            if type(record[first_name]) == unicode:
                                attr_value = str(attr_value.decode("ascii", errors="ignore"))
                            else:
                                attr_value = str(attr_value)

                        try:
                            field_is_string = field_lkp['is_string']
                            if field_is_string == 0:
                                string_field = False
                            else:
                                raise Exception("Incorrect setting for is_string. Value should be number 0")
                        except KeyError:
                            string_field = True

                        if string_field:
                            row += '\"' + attr_value + '\"'
                        else:
                            row += attr_value

                    ''' Add delimiter if not the last field '''
                    if field_lkp['field'] != field_list[-1]['field']:
                        row += file_delimiter

                row += '\n'
                sfdc_file.write(row)

        except KeyError as e:
            print(e, e.args)
            raise Exception("Wrong parameter passed. Check values in attribute_list")

        sfdc_file.close()

        return




import subprocess


def send_email_notification(subject, to, body, attach_files='', attach_names=''):

    attach_list = attach_files.split(',')
    attach_names = attach_names.split(',')

    index = 0
    command = "("
    for attach_file in attach_list:

        try:
            attach_name = attach_names[index]
        except:
            attach_name = attach_file

        command += "uuencode " + attach_file + attach_name + ";"
        index += 1

    command = "echo " + '"' + body + '"' + ") | mailx -s  " + '"' + subject + '" ' + to

    print (command)
    subprocess.call(command, shell=True)

    return



import json
import requests
import sys
import urllib
from base64 import b64encode

maj_ver = sys.version_info.major
if maj_ver == 2:
    import urllib2


class DelightedClient():

    '''
    https://delighted.com/docs/api/
    '''

    API_ENDPOINT = 'https://api.delighted.com'
    VERSION = 'v1'

    def __init__(self, api_credentials):

        self.api_credentials = api_credentials

        ''' Password field is blank. Leave 2nd parameter after : as blank  '''
        encoded_api_key = (self.api_credentials['apikey'] + ":").encode("ascii")

        ''' Pass the api_key as username as a byte object, encode it into base64 and decode it back to ascii '''
        headers = {"Authorization": "Basic " + b64encode(encoded_api_key).decode("ascii")}
        self.api_header = headers

        return

    def unicode_urlencode(self, params):
        """
            Convert lists to JSON encoded strings, and correctly handle any
            unicode URL parameters.
        """
        if isinstance(params , dict):
            params = params.items()
        for i, param in enumerate(params):
            if isinstance(param[1] , list):
                params[i] = (param[0], json.dumps(param[1]),)

        if maj_ver == 2:
            return urllib.urlencode(
                [(k , isinstance(v, unicode) and v.encode('utf-8') or v) for k, v in params]
            )

        if maj_ver == 3:
            return urllib.parse.urlencode(
                [(k, isinstance(v, str) and v.encode('utf-8') or v) for k, v in params]
            )

        return

    def get_metrics(self, params={}, verify=False):

        method_name = "metrics"
        url_endpoint = self.API_ENDPOINT + "/" + self.VERSION + "/" + method_name + ".json"
        my_session = requests.session()

        if params == {}:
            r = my_session.get(url_endpoint, verify=verify, headers=self.api_header)
        else:
            r = my_session.get(url_endpoint + "?" + self.unicode_urlencode(params),
                               verify=verify, headers=self.api_header)

        return r.json()

    def request_data(self, method_name, params={}, verify=False):

        if method_name not in ("unsubscribes", "bounces", "survey_responses"):
            print ("method_name should be 1 of unsubscribes, bounces, metrics or survey_responses")
            raise Exception("Invalid method_name = " + method_name)

        response_list = []
        url_endpoint = self.API_ENDPOINT + "/" + self.VERSION + "/" + method_name + ".json"
        my_session = requests.session()

        i = 0
        more_pages = True
        while more_pages:
            i += 1
            params["page"] = i
            params["per_page"] = 100
            api_response = my_session.get(url_endpoint + "?" + self.unicode_urlencode(params),
                                          verify=verify, headers=self.api_header)

            if len(api_response.json()) > 0:
                for response in api_response.json():
                    response_list.append(response)
            else:
                more_pages = False

        return response_list



import requests
import sys
import urllib
import json

maj_ver = sys.version_info.major
if maj_ver == 2:
    import urllib2


class WootricClient():
    '''
    http://docs.wootric.com/api
    '''

    API_ENDPOINT = 'https://api.wootric.com'
    VERSION = 'v1'

    def __init__(self, api_credentials):
        self.api_credentials = api_credentials

        form = dict(
            grant_type='client_credentials',
            client_id=api_credentials['apikey'],
            client_secret=api_credentials['apisecret']
        )

        resp = requests.post(self.API_ENDPOINT + "/oauth/token", data=form)

        if resp.status_code != 200:
            raise Exception('Unable to authenticate. Check login credentials ')
        else:
            access_token = resp.json().get("access_token", None)
            self.access_token = access_token

        return

    def unicode_urlencode(self, params):
        """
            Convert lists to JSON encoded strings, and correctly handle any
            unicode URL parameters.
        """
        if isinstance(params , dict):
            params = params.items()
        for i, param in enumerate(params):
            if isinstance(param[1] , list):
                params[i] = (param[0], json.dumps(param[1]),)

        if maj_ver == 2:
            return urllib.urlencode(
                [(k , isinstance(v, unicode) and v.encode('utf-8') or v) for k, v in params]
            )

        if maj_ver == 3:
            return urllib.parse.urlencode(
                [(k, isinstance(v, str) and v.encode('utf-8') or v) for k, v in params]
            )

        return

    def request_data(self, method_name, params={}, verify=False):

        url_endpoint = self.API_ENDPOINT + "/" + self.VERSION + "/"

        my_session = requests.session()
        headers = {"Authorization": "Bearer " + self.access_token}
        response_list = []

        if method_name not in ("nps_summary", "responses", "end_users", "declines", "segments", "email_survey"):
            print ("Invalid method_name = " + method_name)
            print ("Check http://docs.wootric.com/api/ for list of avialble methods")
            raise Exception("Invalid method_name = " + method_name)

        i = 0
        more_pages = True
        while more_pages:
            i += 1
            params["page"] = i
            params["per_page"] = 50

            api_response = my_session.get(url_endpoint + method_name + "?" + self.unicode_urlencode(params),
                                          verify=verify, headers=headers)
            try:
                if api_response.json()['message'] == 'Not Found':
                    print("curl request - " + url_endpoint + method_name + "?" + self.unicode_urlencode(params))
                    print("no results returned. check curl request")
                    break
            except:
                if len(api_response.json()) > 0:
                    for response in api_response.json():
                        response_list.append(response)
                else:
                    more_pages = False

                if len(api_response.json()) < 50:
                    more_pages = False

        return response_list



import psycopg2
import sqlalchemy
from sqlalchemy import exc
import time
import datetime as dt
import sys, os
import subprocess
import shutil


class PostgresClient():
    '''
    Connects to greenplum database and executes sql queries against db
    '''

    def __init__(self, config_file, db_credentials, db_host):
        self.config_file = config_file
        self.db_credentials = db_credentials
        self.db_host = db_host
        return

    def get_db_conn(self, use_pyscopg2=True, verbose=False, connect_timeout=10):
        '''
        uses the credentials in config_yaml_file for db_host while initiating database connection
        connection can be established by either using pyscopg2 or sqlachemy modules
        :param use_pyscopg2:
        :param verbose:
        :return: database connection session that has been made using the credentials
        '''

        db_credentials = self.db_credentials
        db_host = self.db_host

        if use_pyscopg2:
            try:
                self.conn = psycopg2.connect(database=db_credentials[db_host]['database'],
                                             user=db_credentials[db_host]['user'],
                                             password=db_credentials[db_host]['password'],
                                             host=db_credentials[db_host]['host'],
                                             port=db_credentials[db_host]['port'],
                                             connect_timeout=connect_timeout)
                if verbose:
                    print("db credentials import succeeded. Connected using psycopg2 ..")

            except KeyError:
                if verbose:
                    print("No password found in yaml file. Using password from pgpass")
                self.conn = psycopg2.connect(database=db_credentials[db_host]['database'],
                                             user=db_credentials[db_host]['user'],
                                             host=db_credentials[db_host]['host'],
                                             port=db_credentials[db_host]['port'],
                                             connect_timeout=connect_timeout)
                if verbose:
                    print("db credentials import succeeded. Connected using psycopg2 ..")

            except psycopg2.Error as e:
                print("Unable to connect using psycopg2!")
                print(e.pgerror)
                print(e.diag.message_detail)
            except Exception as e:
                print(e, e.args)

        else:
            # Using sqlalchemy to make the connection
            try:
                connect_string = "postgresql://" + db_credentials[db_host]['user'] + ":" + \
                                 db_credentials[db_host]['password'] + \
                                 "@" + db_credentials[db_host]['host'] + "/" + \
                                 db_credentials[db_host]['database']
                engine = sqlalchemy.create_engine(connect_string)
                c = engine.connect()
                self.conn = c.connection
                if verbose:
                    print("db credentials import succeeded. connected successfully using sqlalchemy ..")
            except exc.SQLAlchemyError as e:
                print("Unable to connect using sqlalchemy!")
            except Exception as e:
                print(e, e.args)

        return self.conn

    def get_db_cursor(self, verbose=False, connect_timeout=10):
        '''
        uses the credentials in config_yaml_file for db_host while initiating database connection.
        Return the database cursor after making the database connection
        :param verbose:
        :param connect_timeout:
        :return:
        '''

        db_credentials = self.db_credentials
        db_host = self.db_host

        try:
            self.conn = psycopg2.connect(database=db_credentials[db_host]['database'],
                                         user=db_credentials[db_host]['user'],
                                         password=db_credentials[db_host]['password'],
                                         host=db_credentials[db_host]['host'],
                                         port=db_credentials[db_host]['port'],
                                         connect_timeout=connect_timeout)
            if verbose:
                print("db credentials import succeeded. Connected using psycopg2 ..")

        except KeyError:
            if verbose:
                print("No password found in yaml file. Using password from pgpass")
            self.conn = psycopg2.connect(database=db_credentials[db_host]['database'],
                                         user=db_credentials[db_host]['user'],
                                         host=db_credentials[db_host]['host'],
                                         port=db_credentials[db_host]['port'],
                                         connect_timeout=connect_timeout)
            if verbose:
                print("db credentials import succeeded. Connected using psycopg2 ..")

        except psycopg2.Error as e:
            print("Unable to connect using psycopg2!")
            print(e.pgerror)
            print(e.diag.message_detail)
        except Exception as e:
            print(e, e.args)

        try:
            self.cur = self.conn.cursor()
            if verbose:
                print("opened database successfully")
        except Exception as e:
            print("unable to connect to database", "error message: {0}".format(e))

        return self.cur

    def execute_query(self, query):
        try:
            self.cur.execute(query)
        except:
            self.conn.rollback()
            raise
        return None

    def commit_query(self, verbose=False):
        try:
            self.conn.commit()
            if verbose:
                print("successful commit")
        except Exception as e:
            print("error message: {0}".format(e))

    def copy_query(self, query, file_path):
        try:
            f = open(file_path, 'r')
            self.cur.copy_expert(sql=query, file=f)
            self.conn.commit()
            print("copy insert successful")
        except Exception as e:
            print("copy insert error message: {0}".format(e))

    def get_results(self):
        return self.cur.fetchall()

    def rollback(self):
        self.conn.rollback()
        return None

    def close_connection(self):
        self.conn.close()
        return None

    def run_sql_file(self, sql_query_file):
        '''
        The function takes a filename as input
        and will run the SQL query using previously established connection
        pyscopg2 connection must be established prior to calling this function
        :param sql_query_file:
        :return:
        '''

        try:
            f = open(sql_query_file, 'r')
            start = time.time()
            sql = s = " ".join(f.readlines())

            print (
            "Started executing: " + f.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            cursor = self.conn.cursor()
            cursor.execute(sql)
            self.conn.commit()
            cursor.close()

            end = time.time()
            print (
            "Finished executing: " + f.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            print ("Overall run time:")
            print (str((end - start) * 1000) + ' ms')

        except psycopg2.Error as e:
            print ("Unable to connect!")
            print (e.pgerror)
            print (e.diag.message_detail)
        except OSError as err:
            print("OS error: " + format(err))
            cursor.close()
        except Exception as e:
            print("error message: {0}".format(e))
            cursor.close()

        return

    def run_sql_command(self, sql_command, verbose=False, fetch=False):
        '''
        The function takes a sql_coomand and a connection as input and will run the SQL query using given connection
        pyscopg2 connection must be established prior to calling this function.
        fecth=True for a query that returns rows at end of the sql
        :param sql_command:
        :param connection:
        :param fetch:
        :return: results of sql query
        '''

        try:
            start = time.time()
            if verbose:
                print ("Started executing: " + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            cursor = self.conn.cursor()
            if fetch:
                cursor.execute(sql_command)
                results = cursor.fetchall()
            else:
                cursor.execute(sql_command)
                self.conn.commit()
                results = None

            cursor.close()
            end = time.time()
            if verbose:
                print ("Finished executing: " + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
                print ("Overall run time:")
                print (str((end - start) * 1000) + ' ms')

        except psycopg2.Error as e:
            print ("Unable to connect!")
            print (e.pgerror)
            print (e.diag.message_detail)
        except OSError as err:
            print("OS error: " + format(err))
            cursor.close()
        except Exception as e:
            print("error message: {0}".format(e))
            cursor.close()

        return results

    def export_sql_results(self, sql_file, log_file, output_file,
                           output_file_delimited=False, delimiter=','):
        '''
        The function takes a file as input and will run the SQL query using given connection
        pyscopg2 connection must be established prior to calling this function.
        results are redirected to output_file that can be delimited
        :param sql_file:
        :param output_file:
        :param output_file_delimited:
        :param delimiter:
        :return: None
        '''

        try:
            log_out = open(log_file, 'a')

            f = open(sql_file, 'r')
            start = time.time()
            sql = s = " ".join(f.readlines())

            start = time.time()
            print (
            "Started executing: " + f.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            log_out.write(
                "Started executing: " + f.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            cursor = self.conn.cursor()

            if output_file_delimited:
                COPY_STATEMENT = """
                COPY ({sql}) TO STDOUT
                    WITH CSV HEADER
                    DELIMITER AS '{delimiter}'
                """.format(sql=sql, delimiter=delimiter)
            else:
                COPY_STATEMENT = """
                COPY ({sql}) TO STDOUT
                WITH CSV HEADER
                """.format(sql=sql)
            with open(output_file, 'w') as out:
                cursor.copy_expert(COPY_STATEMENT, out)
            cursor.close()

            end = time.time()
            print ("Ended executing: " + f.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            print ("Overall run time:")
            print (str((end - start) * 1000) + ' ms')
            log_out.write(
                "Ended executing: " + f.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            log_out.write("Overall run time:")
            log_out.write(str((end - start) * 1000) + ' ms')

        except psycopg2.Error as e:
            print ("Unable to export!")
            print (e.pgerror)
            print (e.diag.message_detail)
            cursor.close()
        except OSError as err:
            print("OS error: " + format(err))
            cursor.close()
        except Exception as e:
            print("error message: {0}".format(e))
            cursor.close()

        return

    def load_csv_to_table(self, data_file, delimiter, schema_name, table_name, clear_target=False):
        '''
        The function takes input file, file delimiter and a connection as input
        and loads the data into the table. File has to be in csv format and delimited.
        target table must be empty or will be truncated prior to load
        pyscopg2 connection must be established prior to calling this function
        :param data_file:
        :param delimiter:
        :param conn:
        :param schema_name:
        :param table_name:
        :return:
        '''

        try:
            f = open(data_file, 'r')
            cur = self.conn.cursor()
            objectName = schema_name + "." + table_name
            if clear_target:
                query = "truncate " + objectName + ";"
                cur.execute(query)
            SQL_STATEMENT = """
            COPY %s FROM STDIN WITH
                CSV
                HEADER
                DELIMITER AS '{delimiter}'
            """.format(delimiter=delimiter)
            cur.copy_expert(sql=SQL_STATEMENT % objectName, file=f)
            self.conn.commit()
            cur.close()

        except psycopg2.Error as e:
            print ("Unable to load!")
            print (e.pgerror)
            print (e.diag.message_detail)
        except OSError as err:
            print("OS error: " + format(err))
            cur.close()
        except Exception as e:
            print("error message: {0}".format(e))
            cur.close()

        return


class GreenplumClient(PostgresClient):
    def __init__(self, config_file, db_credentials, db_host):
        self.config_file = config_file
        self.db_credentials = db_credentials
        self.db_host = db_host
        return

    def run_psql_file(self, sql_file, log_file, output_file,
                      output_file_delimited=False, delimiter=','):
        '''
        The function takes a sql_file, log file, output_file and format of output_file as input
        and will run the sql queries in the sql file using psql
        psql uses the credentials in config_yaml_file for db_host while initiating database connection
        This module will raise an exception if any ERROR is encountered while running the sql file
        :param sql_file:
        :param log_file:
        :param output_file:
        :param output_file_delimited:
        :return:
        '''

        db_credentials = self.db_credentials
        db_host = self.db_host

        sql_in = open(sql_file, 'r')
        results_out = open(output_file, 'a')
        log_out = open(log_file, 'a')

        try:

            start = time.time()
            print ("Started: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            log_out.write(
                "Started: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            if output_file_delimited:
                psql_string = " psql -h " + db_credentials[db_host]['host'] + " -U " + db_credentials[db_host]['user'] + \
                              " -d " + db_credentials[db_host]['database'] + " -p " + str(
                    db_credentials[db_host]['port']) + \
                              " -f " + os.path.abspath(sql_in.name) + " -A -F '" + delimiter + "' -X  -t -q " + \
                              " > " + os.path.abspath(results_out.name)

            else:
                psql_string = " psql -h " + db_credentials[db_host]['host'] + " -U " + db_credentials[db_host]['user'] + \
                              " -d " + db_credentials[db_host]['database'] + " -p " + str(
                    db_credentials[db_host]['port']) + \
                              " -f " + os.path.abspath(sql_in.name) + " -t > " + os.path.abspath(results_out.name)

            output, error = subprocess.Popen(psql_string, shell=True, stderr=subprocess.STDOUT,
                                              stdout=subprocess.PIPE).communicate()
            if sys.version_info.major == 3:
                output = output.splitlines()
                for line in output:
                    log_out.write(str(line, 'utf-8'))
            else:
                log_out.write(output)

            end = time.time()
            print ("Ended: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")
            log_out.write(
                "Ended: " + sql_in.name + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            total_time = (end - start) * 1000
            print ("Total time taken to run: " + str(total_time) + " ms \n")
            log_out.write("Total time taken to run: " + str(total_time) + " ms \n")

            results_out.close()
            log_out.close()

        except OSError as err:
            print("OS error: {0)".format(err))
            log_out.write("OS error: {0)".format(err))
        except Exception as e:
            print("error message: {0}".format(e))
            log_out.write("error message: {0}".format(e))

        # --------------------------------------------------------------------------------
        # Check the log file for any errors
        log_in = open(log_file, 'r')
        line_no = 0
        for log_line in log_in:
            line_no += 1
            error_position = log_line.find("ERROR:")
            if error_position > 0:
                print ("Error while processing " + sql_in.name)
                print ("Check line number " + str(line_no) + " in " + log_in.name)
                raise Exception('psql execution failed - Error while running sql')
                break
        log_in.close()

        return

    def run_psql_command(self, sql_command, log_file, output_file,
                         output_file_delimited=False, delimiter=','):
        '''
        The function takes a sql_command, log file, output_file and format of output_file as input
        and will run the sql command using psql
        psql uses the credentials in config_yaml_file for db_host while initiating database connection
        This module will raise an exception if any ERROR is encountered while running the sql file
        :param sql_command:
        :param log_file:
        :param output_file:
        :param output_file_delimited:
        :return:
        '''

        db_credentials = self.db_credentials
        db_host = self.db_host

        results_out = open(output_file, 'a')
        log_out = open(log_file, 'a')

        try:
            start = time.time()
            log_out.write("Started: " + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            if output_file_delimited:
                psql_string = " psql -h " + db_credentials[db_host]['host'] + " -U " + db_credentials[db_host]['user'] + \
                              " -d " + db_credentials[db_host]['database'] + " -p " + str(
                    db_credentials[db_host]['port']) + \
                              """ -c " """ + sql_command + """ " -A -F '""" + delimiter + """' -X  -t -q """ + \
                              " > " + os.path.abspath(results_out.name)

            else:
                psql_string = " psql -h " + db_credentials[db_host]['host'] + " -U " + db_credentials[db_host]['user'] + \
                              " -d " + db_credentials[db_host]['database'] + " -p " + str(
                    db_credentials[db_host]['port']) + \
                              """ -c " """ + sql_command + """ " """ + " > " + os.path.abspath(results_out.name)

            output, error = subprocess.Popen(psql_string, shell=True, stderr=subprocess.STDOUT,
                                              stdout=subprocess.PIPE).communicate()
            if sys.version_info.major == 3:
                output = output.splitlines()
                for line in output:
                    log_out.write(str(line, 'utf-8'))
            else:
                log_out.write(output)

            end = time.time()
            log_out.write("Ended: " + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            total_time = (end - start) * 1000
            log_out.write("Total time taken to run: " + str(total_time) + " ms \n")

            results_out.close()
            log_out.close()

        except OSError as err:
            print("OS error: {0}".format(err))
            log_out.write("OS error: {0}".format(err))
        except Exception as e:
            print("error message: {0}".format(e))
            log_out.write("error message: {0}".format(e))

        # --------------------------------------------------------------------------------
        # Check the out file for any errors
        results_in = open(output_file, 'r')
        line_no = 0
        for result_line in results_in:
            line_no += 1
            error_position = result_line.find("ERROR:")
            if error_position > 0:
                print ("Error while processing ")
                print ("Check line number " + str(line_no) + " in " + results_in.name)
                results_in.close()
                raise Exception('psql execution failed - Error while running sql')
                break
        results_in.close()

        return

    def run_gpload_script(self, data_file, gpload_script_location, log_file,
                          gp_target_table, gp_error_table, file_delimiter='|',
                          null_string='', full_refresh=True, clear_target=True,
                          gpload_file_name='',
                          header=True, column_list=[], error_limit="2"):
        '''

        Reference -
        https://gpdb.docs.pivotal.io/510/admin_guide/load/topics/g-loading-data-with-gpload.html

        :param data_file: 
        :param gpload_script_location: 
        :param log_file: 
        :param gp_target_table: 
        :param gp_error_table: 
        :param file_delimiter: 
        :param null_string: 
        :param full_refresh: 
        :param clear_target:
        :param gpload_file_name:
        :param header: 
        :param column_list:
        :param error_limit: 
        :return: 
        '''

        db_credentials = self.db_credentials
        db_host = self.db_host

        if gpload_file_name == '':
            gpload_srcipt_file_name = gpload_script_location + "/gpload_" + gp_target_table + ".yml"
        else:
            gpload_srcipt_file_name = gpload_script_location + "/gpload_" + gpload_file_name + ".yml"

        gpload_srcipt_file = open(gpload_srcipt_file_name, 'w')
        log_out = open(log_file, 'a')

        try:
            start = time.time()
            log_out.write("Started: " + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            gpload_srcipt_file.write("VERSION: 1.0.0.1\n")
            gpload_srcipt_file.write("DATABASE: {0}\n".format(db_credentials[db_host]['database']))
            gpload_srcipt_file.write("USER: {0}\n".format(db_credentials[db_host]['user']))
            gpload_srcipt_file.write("HOST: {0}\n".format(db_credentials[db_host]['host']))
            gpload_srcipt_file.write("PORT: {0}\n".format(db_credentials[db_host]['port']))
            gpload_srcipt_file.write("GPLOAD: \n")
            gpload_srcipt_file.write("   INPUT: \n")
            gpload_srcipt_file.write("      - SOURCE: \n")
            gpload_srcipt_file.write("         FILE: \n")

            if isinstance(data_file, list):
                for file_name in data_file:
                    gpload_srcipt_file.write("             - {0}\n".format(file_name))
            else:
                gpload_srcipt_file.write("             - {0}\n".format(data_file))

            if len(column_list) > 0:
                gpload_srcipt_file.write("      - COLUMNS:\n")
                for column_name in column_list:
                    gpload_srcipt_file.write("              - {0}\n".format(column_name))

            gpload_srcipt_file.write("      - FORMAT: text\n")
            if header:
                gpload_srcipt_file.write("      - HEADER: true\n")
            else:
                gpload_srcipt_file.write("      - HEADER: false\n")
            gpload_srcipt_file.write("      - DELIMITER: '{0}'\n".format(file_delimiter))
            gpload_srcipt_file.write("      - NULL_AS: '{0}'\n".format(null_string))
            gpload_srcipt_file.write("""      - ESCAPE: "OFF"\n""")
            gpload_srcipt_file.write("      - ERROR_LIMIT: {0}\n".format(error_limit))
            gpload_srcipt_file.write("      - ERROR_TABLE: {0}\n".format(gp_error_table))

            gpload_srcipt_file.write("   OUTPUT: \n")
            gpload_srcipt_file.write("       - TABLE: {0}\n".format(gp_target_table))
            gpload_srcipt_file.write("       - MODE: INSERT\n")

            if full_refresh or clear_target:
                gpload_srcipt_file.write("   SQL: \n")

            if full_refresh:
                gpload_srcipt_file.write("""      - BEFORE: "TRUNCATE {0};"\n""".format(gp_target_table))

            if clear_target:
                gpload_srcipt_file.write(
                    """      - AFTER: "ANALYZE {0}; ALTER TABLE {1} SET WITH (reorganize=true)"\n""".
                    format(gp_target_table, gp_target_table))

            gpload_srcipt_file.close()

            gpload_bash_file_name = gpload_script_location + "/gpload_" + gp_target_table + ".sh"
            gpload_bash_file = open(gpload_bash_file_name, 'w')
            gpload_bash_file.write("source " + os.environ['GPLOAD_HOME'] + "/" + "greenplum_loaders_path.sh" + "\n")
            gpload_bash_file.write("gpload -f " + gpload_srcipt_file_name + "\n")
            gpload_bash_file.close()
            gpload_command = "chmod 755 " + gpload_bash_file_name + " && bash " + gpload_bash_file_name
            output, error = subprocess.Popen(gpload_command, shell=True, stderr=subprocess.STDOUT,
                                              stdout=subprocess.PIPE).communicate()

            if sys.version_info.major == 3:
                output = output.splitlines()
                for line in output:
                    log_out.write(str(line, 'utf-8'))
                    print(str(line, 'utf-8'))
            else:
                log_out.write(output)
                print(output)

            end = time.time()
            log_out.write("Ended: " + " at " + str(dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            total_time = (end - start) * 1000
            log_out.write("Total time taken to run: " + str(total_time) + " ms \n")
            log_out.close()

        except OSError as err:
            print("OS error: {0}".format(err))
            log_out.write("OS error: {0}".format(err))
        except Exception as e:
            print("error message: {0}".format(e))
            log_out.write("error message: {0}".format(e))

        # --------------------------------------------------------------------------------
        # Check the log file for any errors
        log_in = open(log_file, 'r')
        line_no = 0
        for log_line in log_in:
            line_no += 1
            error_position = log_line.find("ERROR:")
            if error_position > 0:
                print("Error while processing ")
                print("Check line number " + str(line_no) + " in " + log_in.name)
                log_in.close()
                raise Exception('gpload execution failed - Error while running gpload')
                break
        log_in.close()

        return

    def __remove_line_feeds(self, file_name):

        '''
        
        :param file_name: 
        :return: 
        
        head -n 1 filename | od -c  to find out whats wrong
        
        tr -d '\n' <filename
        tr -d '\r\n' <filename

        '''

        temp_out_file_name = file_name + ".tmp"

        repl_command = """tr -d '\n' < {0}  > {1}""".format(file_name, temp_out_file_name)

        output, error = subprocess.Popen(repl_command, shell=True, stderr=subprocess.STDOUT,
                                         stdout=subprocess.PIPE).communicate()

        if sys.version_info.major == 3:
            output = output.splitlines()
            for line in output:
                print(str(line, 'utf-8'))
        else:
            print(output)

        shutil.move(temp_out_file_name, file_name)



import threading, time
import multiprocessing
import socket
import json
import os
from kafka import KafkaConsumer, KafkaProducer, TopicPartition
from kafka.common import LeaderNotAvailableError
from kafka.errors import KafkaError


'''
                KAFKA BASICS
describe a topic
/usr/local/kafka/bin/kafka-topics.sh --describe --zookeeper localhost:2181 --topic test


creating a topic
/usr/local/kafka/bin/kafka-topics.sh --create --zookeeper localhost:2181 --replication-factor 1 --partitions 2 --topic test
For replication factor to be > 1, there has to be more than 1 kafka broker

deleting a topic
/usr/local/kafka/config
delete.topic.enable=true 

/usr/local/kafka/bin/kafka-topics.sh --zookeeper localhost:2181 --delete --topic test


                python kafka package DOCUMENTATION
http://kafka-python.readthedocs.io/en/master/apidoc/KafkaConsumer.html
http://kafka-python.readthedocs.io/en/master/apidoc/kafka.producer.html

                REFERENCES
http://www.giantflyingsaucer.com/blog/?p=5541
https://pythonexample.com/code/python-kafka-consumer-record/
https://community.hortonworks.com/articles/74077/kafka-producer-sample-code-in-scala-and-python.html
https://www.confluent.io/blog/introduction-to-apache-kafka-for-python-programmers/
http://www.giantflyingsaucer.com/blog/?p=5541
https://github.com/dpkp/kafka-python/blob/master/example.py


                KAFKA Installation
http://davidssysadminnotes.blogspot.in/2016/01/installing-apache-kafka-and-zookeeper.html


                KAFKA GUI Tool
https://github.com/HomeAdvisor/Kafdrop
cd Kafdrop
mvn clean package
java -jar <path to folder>/Kafdrop/target/kafdrop-2.0.0.jar --zookeeper.connect=<zookeeper-host>:2181 --server.port=9090

goto http://localhost:9090/ to see the UI

vi /etc/systemd/system/kafdrop.service

[Unit]
Description=kafkadrop UI for Apache Kakfa server (broker)
Documentation=https://github.com/HomeAdvisor/Kafdrop
Requires=network.target remote-fs.target
After=network.target remote-fs.target zookeeper.service kafka.service

[Service]
Type=simple
User=my_user
Group=my_group
Environment=JAVA_HOME=/opt/jdk1.8.0_121/
ExecStart=/usr/bin/env "${JAVA_HOME}/bin/java" -jar /opt/Kafdrop/target/kafdrop-2.0.0.jar --zookeeper.connect=kakfa-server:2181 --server.port=9090

[Install]
WantedBy=multi-user.target

:wq

systemctl daemon-reload
service kafdrop start

'''


class BatchProducer(threading.Thread):
    def __init__(self, config_file, db_credentials, db_host):
        threading.Thread.__init__(self)
        self.config_file = config_file
        self.stop_event = threading.Event()
        self.db_credentials = db_credentials
        self.db_host = db_host
        self.bootstrap_servers = self.db_credentials[self.db_host]['bootstrap_servers']
        return

    def stop(self):
        self.stop_event.set()
        return

    def send(self, topic_name, message, topic_partition=-1):

        producer = KafkaProducer(bootstrap_servers=self.bootstrap_servers,
                                 value_serializer=lambda v: json.dumps(v).encode('utf-8'))

        try:
            exception = ""
            if topic_partition == -1:
                producer.send(topic_name, message)
            else:
                producer.send(topic_name, message, partition=topic_partition)

        except LeaderNotAvailableError:
            # https://github.com/mumrah/kafka-python/issues/249
            time.sleep(1)
            if topic_partition == -1:
                producer.send(topic_name, message)
            else:
                producer.send(topic_name, message, partition=topic_partition)

        except KafkaError as ex:
            print(ex)
            exception = str(ex)

        finally:
            producer.close()

        return exception


class BatchConsumer(multiprocessing.Process):

    def __init__(self, config_file, db_credentials, db_host):

        multiprocessing.Process.__init__(self)
        self.stop_event = multiprocessing.Event()
        self.config_file = config_file
        self.db_credentials = db_credentials
        self.db_host = db_host
        self.bootstrap_servers = self.db_credentials[self.db_host]['bootstrap_servers']

        return

    def stop(self):
        self.stop_event.set()
        return

    def connect(self, group_id="", client_id="",
                auto_offset_reset='earliest', consumer_timeout_ms=1000):

        client_id = socket.gethostname() if client_id == "" else client_id
        group_id = os.getlogin() if group_id == "" else group_id

        exception = ""
        try:
            consumer = KafkaConsumer(bootstrap_servers=self.bootstrap_servers,
                                     auto_offset_reset=auto_offset_reset,
                                     consumer_timeout_ms=consumer_timeout_ms,
                                     value_deserializer=lambda m: json.loads(m.decode('utf-8')),
                                     group_id=group_id,
                                     client_id=client_id,
                                     enable_auto_commit=False
                                     )
            self.consumer = consumer

        except KafkaError as ex:
            print(ex)
            exception = str(ex)
            self.consumer = None

        return exception

    def receive(self, topic_name, topic_partition=None, message_offset=None,
                out_dir='', out_file_name=''):

        exception = ""
        consumer = self.consumer

        try:
            if out_dir == '':
                out_dir = os.environ['HOME']

            if topic_partition is None:

                ''' Get all messages from all partitions '''
                consumer.subscribe(topic_name)

                if out_file_name == '':
                    out_file_name = str(self.db_host) + "_" + str(topic_name)
                out_file = open(out_dir + "/" + out_file_name + "_all.txt", 'w')

                for message in consumer:
                    # This will wait and print messages as they become available
                    kafka_message = dict()
                    kafka_message["topic"] = message.topic
                    kafka_message["partition"] = message.partition
                    kafka_message["offset"] = message.offset
                    kafka_message["key"] = message.key
                    kafka_message["value"] = message.value
                    kafka_message["timestamp"] = message.timestamp
                    kafka_message["timestamp_type"] = message.timestamp_type
                    out_file.write(json.dumps(kafka_message) + "\n")
                out_file.close()

            else:
                partition_list = [TopicPartition(topic_name, topic_partition)]
                consumer.assign(partition_list)

                ''' Start fetching data from partition '''
                for topic_partition in consumer.assignment():
                    print("fetching results for {0} ...".format(topic_partition))

                    if message_offset is not None:
                        print("setting offset to {0}".format(message_offset))
                        consumer.seek(partition=topic_partition, offset=message_offset)
                    else:
                        print("last committed offset {0}".format(consumer.end_offsets([topic_partition])[topic_partition]))
                        consumer.poll()

                    if out_file_name == '':
                        out_file_name = str(self.db_host) + "_" + str(topic_name)

                    if out_dir == '':
                        out_dir = os.environ['HOME']
                    out_file = open(out_dir + "/" + out_file_name + "_" + str(topic_partition.partition) + ".txt", 'w')
                    for message in consumer:
                        # This will wait and print messages as they become available
                        kafka_message = dict()
                        kafka_message["topic"] = message.topic
                        kafka_message["partition"] = message.partition
                        kafka_message["offset"] = message.offset
                        kafka_message["key"] = message.key
                        kafka_message["value"] = message.value
                        kafka_message["timestamp"] = message.timestamp
                        kafka_message["timestamp_type"] = message.timestamp_type
                        out_file.write(json.dumps(kafka_message) + "\n")
                    out_file.close()

        except KafkaError as ex:
            print(ex)
            exception = str(ex)
            consumer.close()
        except KeyboardInterrupt:
            pass

        return exception

    def commit(self):
        self.consumer.commit()
        return

    def close(self):
        self.consumer.close()
        return




import redis
import csv


class RedisClient():
    '''
    Connects to redis cache and loads data to a cache
    '''
    def __init__(self, config_file, db_credentials, db_host, debug=False):
        self.config_file = config_file
        self.db_credentials = db_credentials
        self.db_host = db_host
        self.debug = debug
        return

    def get_db_conn(self, other_db=0):
        redis_credentials = self.db_credentials[self.db_host]
        try:
            if other_db == 0:
                conn = redis.Redis(redis_credentials['host'],
                                   redis_credentials['port'],
                                   redis_credentials['database']
                                   )
            else:
                conn = redis.Redis(redis_credentials['host'],
                                   redis_credentials['port'],
                                   other_db
                                   )
        except redis.Error as e:
            print("Unable to connect using redis!")
            print(e.diag.message_detail)
        except Exception as e:
            print(e, e.args)
        return conn

    def __load_data_from_csv(self, csv_file, file_delimiter, ik, iv):
        with open(csv_file) as csvf:
            csv_data = csv.reader(csvf, delimiter=file_delimiter)
            return [(r[ik], r[iv]) for r in csv_data]

    def __store_data(self, conn, data):
        for i in data:
            conn.setnx(i[0], i[1])
        return

    def clear_redis_data(self, conn, clear_all=False):
        if clear_all:
            conn.flushall()
        else:
            conn.flushdb()
        return

    def get_redis_data(self, conn, lookup_key):
        lookup_value = conn.get(lookup_key)
        if lookup_value is None:
            return 'Not-Found'
        else:
            return lookup_value

    def load_csv_to_redis(self, conn, file_name, key_position,
                          value_position, csv_delimiter=','):
        data = self.__load_data_from_csv(file_name, csv_delimiter, key_position,
                                         value_position)
        self.__store_data(conn, data)
        return



import pika
import threading
import os


class Producer(threading.Thread):
    def __init__(self, config_file, db_credentials, db_host):

        threading.Thread.__init__(self)
        self.stop_event = threading.Event()

        self.config_file = config_file
        self.db_credentials = db_credentials
        self.db_host = db_host
        return

    def stop(self):
        self.stop_event.set()
        return

    def send(self, queue_name, message, auto_delete=False):

        db_credentials = self.db_credentials
        rabbitmq_host = self.db_host
        exception = ""

        connect_string = 'amqp://' + db_credentials[rabbitmq_host]['user'] + ':' + \
                         db_credentials[rabbitmq_host]['password'] + '@' + \
                         db_credentials[rabbitmq_host]['host'] + ':' + \
                         str(db_credentials[rabbitmq_host]['port']) + '/' + \
                         db_credentials[rabbitmq_host]['vhost'] + '/%2f'

        url = os.environ.get('CLOUDAMQP_URL', connect_string)
        try:
            params = pika.URLParameters(url)
            params.socket_timeout = 5
            connection = pika.BlockingConnection(params)  # Connect to RabbitMQ Host
        except Exception as e:
            exception = "Unable to connect to rabbitmq. error message: {0}".format(e)

        if exception == "":
            try:
                channel = connection.channel()  # start a channel
                channel.queue_declare(queue=queue_name , durable=True ,
                                      auto_delete=auto_delete)  # Declare a queue - durable=True
                channel.basic_publish(exchange='' ,
                                      routing_key=queue_name,
                                      body=message ,
                                      properties=pika.BasicProperties(
                                          delivery_mode=2,  # make message persistent
                                        )
                                      )
                connection.close()
            except Exception as e:
                connection.close()
                exception = "Error while transfer meesage. error message: {0}".format(e)

        return exception



import base64
import json
import urllib
import sys
import subprocess
import requests

maj_ver = sys.version_info.major
if maj_ver == 2:
    import urllib2


class MixpanelClient():

    API_ENDPOINT = 'https://mixpanel.com/api'
    EXPORT_ENDPOINT = 'https://data.mixpanel.com/api'
    VERSION = '2.0'

    def __init__(self, api_credentials):
        self.api_credentials = api_credentials

    def request_data(self, methods, params={}, output_file_name='',
                     secure=False, debug=False, http_method="GET", timeout=1200):
        '''
        
        :param methods: List of methods to be joined, e.g. ['events', 'properties', 'values']
                        will give us https://mixpanel.com/api/2.0/events/properties/values/
                        
                        if method is export then 
                        https://data.mixpanel.com/api/2.0/export params
        :param params:  Extra parameters associated with method
        :param output_file_name: 
        :param secure: secure - True: verifies ssl certificate else check is bypassed
        :param debug: debug - False: don't print interim results
        :param http_method: GET by default
        :param timeout: 1200 by default
        :return: 
        '''

        mixpanel_credentials = self.api_credentials
        api_secret = mixpanel_credentials['apisecret']

        if methods[0] == 'export' and len(methods) == 1:
            ''' export api request '''
            request_url = '/'.join([self.EXPORT_ENDPOINT, str(self.VERSION)] + methods)
        else:
            request_url = '/'.join([self.API_ENDPOINT, str(self.VERSION)] + methods)

        if http_method == 'GET':
            data = None
            request_url = request_url + '/?' + self.unicode_urlencode(params)
            if debug:
                print (request_url)
        else:
            data = self.unicode_urlencode(params)

        if maj_ver == 2:
            headers = {'Authorization': 'Basic {encoded_secret}'.format(encoded_secret=base64.b64encode(api_secret))}

        if maj_ver == 3:
            headers = {'Authorization': 'Basic {encoded_secret}'.format(encoded_secret=api_secret)}

        resp = requests.get(request_url, verify=secure, headers=headers, timeout=timeout)

        if debug:
            print ("text : {0}".format(resp.text.encode('utf-8')))
            print ("reason : {0}".format(resp.reason.encode('utf-8')))

        if resp.status_code == 200:
            if "export terminated early" in resp.text.encode('utf-8'):
                print ("mixpanel API returned warning - {0}".format(resp.text))
                print ("try again later or contact mixpanel support at support@mixpanel.com")
                raise Warning("unable to export events from mixpanel")
            else:
                if output_file_name == '':
                    print ("No output file name found in batch mode")
                else:
                    resp_data = resp.text.split('\n')
                    with open(output_file_name, 'w') as out:
                        for row in resp_data:
                            out.write(row.encode('utf-8') + '\n')
                        out.close
        else:
            print("Response from API - {0}".format(resp.text))
            print("Error Reason - {0}".format(resp.reason))
            print("HTTP Status Code {0}".format(resp.status_code))
            raise Exception("Mixpanel API Request failed")

    def stream_data(self, methods, params={}, secure=False, debug=False,
                    http_method="GET", timeout=1200):
        '''
        :param methods: List of methods to be joined, e.g. ['events', 'properties', 'values']
                        will give us https://mixpanel.com/api/2.0/events/properties/values/

                        if method is export then 
                        https://data.mixpanel.com/api/2.0/export params
        :param params:  Extra parameters associated with method
        :param secure: secure - True: verifies ssl certificate else check is bypassed
        :param debug: debug - False: don't print interim results
        :param http_method: GET by default
        :param timeout: 1200 by default
        :return: 
        '''

        mixpanel_credentials = self.api_credentials
        api_secret = mixpanel_credentials['apisecret']

        if methods[0] == 'export' and len(methods) == 1:
            ''' export api request '''
            request_url = '/'.join([self.EXPORT_ENDPOINT , str(self.VERSION)] + methods)
        else:
            request_url = '/'.join([self.API_ENDPOINT , str(self.VERSION)] + methods)

        if http_method == 'GET':
            data = None
            request_url = request_url + '/?' + self.unicode_urlencode(params)
            if debug:
                print (request_url)
        else:
            data = self.unicode_urlencode(params)

        if maj_ver == 2:
            headers = {'Authorization': 'Basic {encoded_secret}'.format(encoded_secret=base64.b64encode(api_secret))}

        if maj_ver == 3:
            headers = {'Authorization': 'Basic {encoded_secret}'.format(encoded_secret=api_secret)}

        resp = requests.get(request_url , verify=secure, headers=headers, timeout=timeout, stream=True)

        if debug:
            print ("text : {0}".format(resp.text.encode('utf-8')))
            print ("reason : {0}".format(resp.reason.encode('utf-8')))

        if resp.status_code == 200:
            if "export terminated early" in resp.text.encode('utf-8'):
                print ("mixpanel API returned warning - {0}".format(resp.text))
                print ("try again later or contact mixpanel support at support@mixpanel.com")
                raise Warning("unable to export events from mixpanel")
            else:
                buffer = ""
                for chunk in resp.iter_content(chunk_size=1):
                    if chunk.endswith("\n"):
                        buffer += chunk
                        yield buffer
                        buffer = ""
                    else:
                        buffer += chunk

        else:
            print("Response from API - {0}".format(resp.text))
            print("Error Reason - {0}".format(resp.reason))
            print("HTTP Status Code {0}".format(resp.status_code))
            raise Exception("Mixpanel API Request failed")

    def unicode_urlencode(self, params):
        """
            Convert lists to JSON encoded strings, and correctly handle any
            unicode URL parameters.
        """
        if isinstance(params, dict):
            params = params.items()
        for i, param in enumerate(params):
            if isinstance(param[1] , list):
                params[i] = (param[0], json.dumps(param[1]),)

        if maj_ver == 2:
            return urllib.urlencode(
                [(k, isinstance(v, unicode) and v.encode('utf-8') or v) for k, v in params]
            )

        if maj_ver == 3:
            return urllib.parse.urlencode(
                [(k, isinstance(v, str) and v.encode('utf-8') or v) for k, v in params]
            )

        return

    def run_mixpanel_query_file(self, script_file, output_file, secure=False, pretty=False, debug=False):

        mixpanel_credentials = self.api_credentials
        api_secret = mixpanel_credentials['apisecret']

        if secure:
            secure_flag = ""
        else:
            secure_flag = "--insecure"

        if pretty:
            pretty_format = ""
        else:
            pretty_format = " | python -m json.tool"

        url_endpoint = '/'.join([self.API_ENDPOINT, str(self.VERSION), 'jql'])

        rest_api_call = \
            """curl {URL} {secure_flag} -u {secret_key}: --data-urlencode script@{script_file} \
            {pretty_str} > {output_file}""".format(URL=url_endpoint, secret_key=api_secret,
                                                   pretty_str=pretty_format,
                                                   secure_flag=secure_flag,
                                                   script_file=script_file,
                                                   output_file=output_file)
        if debug:
            print(rest_api_call)

        output, error = subprocess.Popen(rest_api_call, shell=True, stderr=subprocess.STDOUT,
                                         stdout=subprocess.PIPE).communicate()
        if sys.version_info.major == 3:
            output = output.splitlines()
            for line in output:
                print(str(line, 'utf-8'))
        else:
            print(output)
        return





"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
# To use a consistent encoding
from codecs import open
from os import path
from pkg_resources import require, DistributionNotFound, VersionConflict
import sys

maj_ver = sys.version_info.major

if maj_ver == 3:
    from setuptools import setup, find_packages
else:
    from distutils.core import setup

try:
    from pypandoc import convert
    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("warning: pypandoc module not found, could not convert Markdown to RST")
    read_md = lambda f: open(f, 'r').read()

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
long_description=read_md('README.md')

try:
    require('ConflictingDistribution')
    print()
    print ('You have ConflictingDistribution installed.')
    print ( 'You need to remove ConflictingDistribution from your site-packages')
    print ( 'before installing this software, or conflicts may result.')
    print ()
    sys.exit()

except (DistributionNotFound, VersionConflict):
    pass

if maj_ver == 3:
    setup(
        name='dattasa',
        packages=find_packages(),
        # Versions should comply with PEP440.  For a discussion on single-sourcing
        # the version across setup3.py and the project code, see
        # https://packaging.python.org/en/latest/single_source_version.html
        version='1.1',

        description='Python wrapper for connecting to postgres/greenplum, mysql, mongodb, kafka, redis, mixpanel and salesforce.\
        Also included are modules for performing secure file transfer and sourcing environment variables.',
        long_description=long_description,

        # The project's main homepage.
        url='https://github.com/kartikra/dattasa',

        # Author details
        author='Kartik Ramasubramanian',
        author_email='r.kartik@berkeley.edu',

        # Choose your license
        license='GNU',

        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 4 - Beta',

            # Indicate who your project is intended for

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: MIT License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
        ],

        # What does your project relate to?
        keywords='greenplum postgres mysql mongodb kafka redis rabbitmq salesforce mixpanel delighted wootric data pipeline',

        # You can just specify the packages manually here if your project is
        # simple. Or you can use find_packages().
        # packages=find_packages(exclude=['contrib', 'docs', 'tests']),

        # Alternatively, if you want to distribute just a my_module.py, uncomment
        # this:
        #   py_modules=["my_module"],

        # List run-time dependencies here.  These will be installed by pip when
        # your project is installed. For an analysis of "install_requires" vs pip's
        # requirements files see:
        # https://packaging.python.org/en/latest/requirements.html
        install_requires=['paramiko', 'pyyaml', 'simple_salesforce', 'pymongo', 'python-gnupg',
                          'psycopg2', 'PyMySQL', 'sqlalchemy', 'redis', 'pika', 'kafka', 'kafka-python'],

        # List additional groups of dependencies here (e.g. development
        # dependencies). You can install these using the following syntax,
        # for example:
        # $ pip install -e .[dev,test]

    )

else:
    setup(
        name='dattasa',
        version='1.1',
        packages=['dattasa'],
        url='https://github.com/kartikra/dattasa',
        license='MIT',
        author='Kartik Ramasubramanian',
        author_email='r.kartik@berkeley.edu',
        description='Python wrapper for connecting to postgres/greenplum, mysql, mongodb, kafka, redis, mixpanel  salesforce \
        delighted and wootric. Also included are modules for performing secure file transfer and sourcing environment variables.'
    )



# data_pipeline.py
import datetime as dt
import glob
import multiprocessing
import os
import shutil
import time
import yaml
from dattasa import mixpanel_client
from dattasa import salesforce_client
from dattasa import wootric_client
from dattasa import delighted_client
from dattasa import mysql_client
from dattasa import postgres_client
from dattasa import mongo_client
from dattasa import redis_client
from dattasa import kafka_system
from dattasa import rabbitmq_system


class DataComponent():
    '''
    Each data pipeline can comprise of 1 or more of the following methods -
    get_db_conn
    get_db_cursor
    execute_query           run a sql that does not return any rows
    get_results             return results as a dataframe
    commit_query
    rollback
    close_connection
    run_sql_file            run sql from a file and fetch final result if needed
    run_sql_command         run sql command and fetch final result if needed
    export_sql_results      using mysql/psql command
    load_csv_to_table       load file using COPY command
    run_gpload_script       (only for greenplum)
    '''

    def __init__(self):
        self.config_file = os.environ['HOME'] + "/database.yaml"
        return

    def set_credentials(self, config_db, db_config_file=''):
        try:
            from yaml import CLoader as Loader, CDumper as Dumper
        except ImportError:
            from yaml import Loader, Dumper

        if db_config_file == '':
            db_config_file = self.config_file

        with open(db_config_file, 'r') as f:
            db_credentials = yaml.load(f, Loader=Loader)

        try:
            database_adapter = db_credentials[config_db]['adapter']
            ''' Database Connections '''
            if database_adapter == "mysql":
                return mysql_client.MySQLClient(db_config_file, db_credentials, config_db)
            if database_adapter == "postgres":
                return postgres_client.PostgresClient(db_config_file, db_credentials, config_db)
            if database_adapter == "greenplum":
                return postgres_client.GreenplumClient(db_config_file, db_credentials, config_db)
            if database_adapter == "mongodb":
                return mongo_client.MongoClient(db_config_file, db_credentials, config_db)
            if database_adapter == "redis":
                return redis_client.RedisClient(db_config_file, db_credentials, config_db)
            if database_adapter == "kafka-batch":
                return kafka_system.BatchProducer(db_config_file, db_credentials, config_db),\
                       kafka_system.BatchConsumer(db_config_file, db_credentials, config_db)
            if database_adapter == "rabbit":
                return rabbitmq_system.Producer(db_config_file, db_credentials, config_db)
        except KeyError:
            raise Exception('database adapter not found for ' + config_db + '.. check config file')

        return


class APICall():
    '''
       Each APICall  can comprise of 1 or more of the following methods -
       request_data
       update_data
       unicode_urlencode
       get_salesforce_client  (only for Salesforce)
    '''
    def __init__(self):
        self.config_file = os.environ['HOME'] + "/database.yaml"
        return

    def set_credentials(self, config_db, db_config_file=''):
        try:
            from yaml import CLoader as Loader, CDumper as Dumper
        except ImportError:
            from yaml import Loader, Dumper

        if db_config_file == '':
            db_config_file = self.config_file

        with open(db_config_file, 'r') as f:
            db_credentials = yaml.load(f, Loader=Loader)

        try:
            database_adapter = db_credentials[config_db]['adapter']
            ''' API Connections '''
            if database_adapter == "api":
                api_credentials = db_credentials[config_db]
                api_service_name = config_db
                try:
                    api_service_provider = api_credentials['service']
                    if api_service_provider == "salesforce":
                        return salesforce_client.SalesforceClient(api_credentials)
                    if api_service_provider == "mixpanel":
                        return mixpanel_client.MixpanelClient(api_credentials)
                    if api_service_provider == "wootric":
                        return wootric_client.WootricClient(api_credentials)
                    if api_service_provider == "delighted":
                        return delighted_client.DelightedClient(api_credentials)
                except KeyError:
                    raise Exception('service not found for ' + api_service_name + '.. check scripts file')
            else:
                raise Exception('check value of adapter for ' + config_db + '.. check scripts file')

        except KeyError:
            raise Exception('database adapter not found for ' + config_db + '.. check scripts file')

        return


class DataProcessor():
    '''
    
    '''
    def __init__(self, source_conn, target_conn='', config_file='',
                 source_file='', target_file='', temp_dir='',
                 load_parameters={}):

        self.source_conn = source_conn
        self.target_conn = target_conn
        self.source_file = source_file
        self.target_file = target_file
        self.load_parameters = load_parameters

        if temp_dir == '':
            self.temp_dir = os.environ['TEMP_DIR']
        else:
            self.temp_dir = temp_dir

        if config_file == '':
            self.config_file = os.environ['HOME'] + '/database.yaml'
        else:
            self.config_file = config_file

        try:
            from yaml import CLoader as Loader, CDumper as Dumper
        except ImportError:
            from yaml import Loader, Dumper

        db_config_file = self.config_file
        with open(db_config_file, 'r') as f:
            db_credentials = yaml.load(f, Loader=Loader)
            self.db_credentials = db_credentials

        return

    def multiprocess_sql_query_run(self, sql_file, log_file, out_file, sql_parameters_list,
                                   delimited_file=False, file_delimiter=','):
        '''
        While creating your sql file make sure to replace all parameters p[parameter_name]
        sql_parameter_list comprises of key value pairs in the form {"parameter_name": "parameter_value" }
        :param sql_file: 
        :param log_file: 
        :param out_file: 
        :param sql_parameters_list: 
        :param delimited_file: 
        :param file_delimiter: 
        :return: 
        '''
        source_adapter = self.db_credentials[self.source_conn]['adapter']
        data_source = DataComponent().set_credentials(self.source_conn, self.config_file)
        ''' Check if connection can be made '''
        data_source.get_db_conn()
        data_source.close_connection()

        if source_adapter in ('greenplum', 'postgres', 'mysql'):
            '''
            Use Multiprocessing - For each subset, substitute varaibles in the sql 
            and run each sql as a separate process
            '''
            jobs = []
            sql_file_name = os.path.basename(sql_file)
            file_name = sql_file_name.lower().replace('.sql', '')

            temp_dir = self.temp_dir
            log_out = open(log_file, 'a')

            f = open(sql_file, 'r')
            sql_query = s = " ".join(f.readlines())

            ''' Delete temporary files '''
            for temp_sql_file in glob.glob(temp_dir + file_name + '*.sql'):
                os.remove(temp_sql_file)
            for temp_log_file in glob.glob(temp_dir + file_name + '*.log'):
                os.remove(temp_log_file)
            for temp_out_file in glob.glob(temp_dir + file_name + '*.out'):
                os.remove(temp_out_file)

            file_index = 0
            for sql_parameters in sql_parameters_list:
                file_index += 1
                sql_log_file = temp_dir + file_name + '_' + str(file_index) + '.log'
                sql_out_file = temp_dir + file_name + '_' + str(file_index) + '.out'
                temp_query = sql_query.format(p=sql_parameters)
                sql_query_file = temp_dir + file_name + '_' + str(file_index) + '.sql'
                with open(sql_query_file, 'w') as f:
                    f.write(temp_query)
                    f.close
                p = multiprocessing.Process(target=run_sql_process,
                                            args=(self.source_conn,self.config_file, source_adapter,
                                                  sql_query_file, sql_log_file, sql_out_file,
                                                  delimited_file, file_delimiter))
                jobs.append(p)
                p.start()
                time.sleep(1)
            '''
            Since multiprocessing is used, we want to make sure that every process completed with exit code=0
            '''
            job_exit_codes = []
            for j in jobs:
                j.join()
                job_exit_codes.append(j.exitcode)
                log_out.write("Exiting " + j.name + " with exitcode = " + str(j.exitcode) + "\n")
                log_out.write(
                    "Completed running " + j.name + " at " + str(
                        dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")) + "\n")

            ''' Raise an exception if any exit code is non zero '''
            if any(job_exit_codes) != 0:
                log_out.write('psql execution failed while running ' + sql_file + '\n')
                raise Exception('psql execution failed while running ' + sql_file)

            ''' Delete temporary sql files '''
            for temp_sql_file in glob.glob(temp_dir + file_name + '*.sql'):
                os.remove(temp_sql_file)

            ''' Merge temporary log files '''
            log_out.close()
            log_out = open(log_file, 'ab')
            for log_file in glob.glob(temp_dir + file_name + '*.log'):
                shutil.copyfileobj(open(log_file, 'rb'), log_out)
                os.remove(log_file)
            log_out.close()

            ''' Merge temporary out files '''
            results_out = open(out_file, 'wb')
            for results_file in glob.glob(temp_dir + file_name + '*.out'):
                shutil.copyfileobj(open(results_file, 'rb'), results_out)
                os.remove(results_file)
            results_out.close()

        return

    def bulk_load_source_table_to_gp(self, source_table, load_log_file, target_table, target_err_table="",
                                     file_delimiter='|', source_extract_sql=None, full_refresh=True,
                                     clear_target=True, csv_header=False, null_string='', error_limit="2"):

        source_adapter = self.db_credentials[self.source_conn]['adapter']
        if source_adapter in ('greenplum', 'postgres', 'mysql'):

            if source_extract_sql:
                sql_file = source_extract_sql
                file_name = os.path.basename(source_extract_sql)
                file_name.replace('.sql', '')
                extract_file = self.temp_dir + file_name + ".csv"
            else:
                sql_file = self.temp_dir + source_table + ".sql"
                sql_out = open(sql_file, 'w')
                sql_out.write("SELECT * FROM " + source_table)
                sql_out.close()
                extract_file = self.temp_dir + source_table + ".csv"

            if source_adapter == 'mysql':
                data_mysql = DataComponent().set_credentials(self.source_conn, self.config_file)
                data_mysql.get_db_conn()
                data_mysql.export_sql_results(sql_file, load_log_file, extract_file, True, file_delimiter)
            elif source_adapter == 'postgres':
                data_postgres = DataComponent().set_credentials(self.source_conn, self.config_file)
                data_postgres.get_db_conn()
                data_postgres.export_sql_results(sql_file, load_log_file, extract_file, True, file_delimiter)
            elif source_adapter == 'greenplum':
                data_gp = DataComponent().set_credentials(self.source_conn, self.config_file)
                data_gp.get_db_conn()
                data_gp.run_psql_file(sql_file, load_log_file, extract_file, True, file_delimiter)
        else:
            raise Exception('postgres, mysql or greenplum source needed to use this method')

        target_adapter = self.db_credentials[self.target_conn]['adapter']
        if target_adapter == 'greenplum':
            if target_err_table == "":
                target_err_table = target_table + "_err"
            gpload_script_location = self.temp_dir
            gpload_file_name = "load_" + self.target_conn + "_" + target_table
            column_list = []

            data_gp = DataComponent().set_credentials(self.target_conn, self.config_file)
            data_gp.get_db_conn()
            data_gp.run_gpload_script(extract_file, gpload_script_location, load_log_file,
                                      target_table, target_err_table, file_delimiter, null_string,
                                      full_refresh, clear_target, gpload_file_name,
                                      csv_header, column_list, error_limit)
        else:
            raise Exception('greenplum target needed to use this method')

        return


def run_sql_process(source_conn, config_file, source_adapter, sql_query_file, log_file,
                    out_file, delimited_file=False, delimiter=","):
    '''
    Keeping this method outside of DataProcessor class to enable parallel processing.
    For multiprocessing, each function call must be independent
    :param source_conn: 
    :param config_file: 
    :param source_adapter: 
    :param sql_query_file: 
    :param out_file: 
    :param log_file: 
    :param delimited_file: 
    :param delimiter: 
    :return: 
    '''

    data_comp = DataComponent().set_credentials(source_conn, config_file)

    if source_adapter == 'greenplum':
        data_comp.run_psql_file(sql_query_file, log_file, out_file, delimited_file, delimiter)
    else:
        data_comp.get_db_conn()
        data_comp.export_sql_results(sql_query_file, log_file, out_file,delimited_file, delimiter)
        data_comp.commit_query()
        data_comp.close_connection()

    return



import os, sys, subprocess, json
import pprint

def get_os_env(env_file=os.environ['HOME'] + "/.bashrc"):
    '''
    Return all the os variables in the environment profile file
    :param env_file: Specify the profile file from where to source. Default=$HOME/.bashrc
    :return:
    '''
    source = "source " + env_file
    dump = '/usr/bin/python -c "import os, json;print json.dumps(dict(os.environ))"'
    pipe = subprocess.Popen(['/bin/bash', '-c', '%s && %s' %(source,dump)], stdout=subprocess.PIPE)
    if sys.version_info.major == 3:
        env = json.loads(str(pipe.stdout.read(),'utf-8'))
    else:
        env = json.loads(pipe.stdout.read())

    return env


if __name__ == '__main__':
    my_env = get_os_env()
    pprint.pprint(my_env)
